import unittest
import os
import pandas as pd    
from sklearn import svm
from sklearn.model_selection import train_test_split

from aix360.algorithms.ted import TED_CartesianExplainer
from aix360.datasets.ted_dataset import TEDDataset

class TestTED_Cartesian(unittest.TestCase):
   """
   A class to test the TED_Cartesian explainer
   """

   def setUp(self):
      """
      Perform setup code of opening datafile, constructing train/test split, created TED_CartesianExplainer and
      """

      # Decompose the dataset into X, Y, E     
      X, Y, E = TEDDataset().load_file('Retention.csv')               

      # set up train/test split
      self.X_train, self.X_test, self.Y_train, self.Y_test, self.E_train, self.E_test = train_test_split(X, Y, E, test_size=0.20, random_state=0)

      # Create classifier and pass to TED_Cartesian
      self.estimator = svm.SVC(kernel='linear')   
      self.ted = TED_CartesianExplainer(self.estimator)
      self.ted.fit(self.X_train, self.Y_train, self.E_train)   # train classifier

   def test_instances(self):
      """
      Test 2 instances from the test dataset (hard-coded here) to ensure they give the correct Y and E value.
      Test all 3 interfaces to TED_Cartesian:  predict_explain, predict, and explain
      """

      X1 = [[1, 2, -11, -3, -2, -2,  22, 22]]
      # correct answers:  Y:0; E:3
      Y, E = self.ted.predict_explain(X1)
      self.assertEqual(Y, -10)
      self.assertEqual(E, 13)

      Y = self.ted.predict(X1)
      self.assertEqual(Y, -10)

      E = self.ted.explain(X1)
      self.assertEqual(E, 13)

      X2 = [[3, 1, -11, -2, -2, -2, 296, 0]]
      ## correct answers: Y:-11, E:25
      Y, E = self.ted.predict_explain(X2)
      self.assertEqual(Y, -11)
      self.assertEqual(E, 25)

      Y = self.ted.predict(X2)
      self.assertEqual(Y, -11)

      E = self.ted.explain(X2)
      self.assertEqual(E, 25)

   def test_score(self): 
      """
      This test evaluates the accuracy (in 3 ways) of the TED-enhanced classifier on the full test suite.
      It ensures the accuracy is above the correct values for this classifier.
      NOTE: accuracy will depend on the classifier used.
      """
      YE_accuracy, Y_accuracy, E_accuracy = self.ted.score(self.X_test, self.Y_test, self.E_test)    # evaluate the classifier

      # Expected Accuracy for this dataset and classifier
      # Accuracy of YE = 0.8510
      # Accuracy of Y = 0.8615
      # Accuracy of E = 0.8510

      self.assertGreaterEqual(YE_accuracy, 0.85)
      self.assertGreaterEqual(Y_accuracy, 0.85)
      self.assertGreaterEqual(E_accuracy, 0.85)

if __name__ == '__main__':
    unittest.main()



import os
import requests

class dwnld_CEM_MAF_celebA():
    '''
    Class with functions to download
        1. celebA prediction model
        2. celebA attribute functions
        3. celebA data files
    '''
        
    def dwnld_celebA_attributes(self, local_path, attributes):
        '''
        Download celebA attribute functions
        
        Args:
            local path (str): local path to where files are downloaded
            attributes (str list): list of attributes to download attribute functions for 

        Returns:
            files (str list): list of files that were downloaded
        '''

        # This is the link where attribute functions are stored
        cdcweb = 'http://aix360-dev.mybluemix.net/static/CEM-MAF/attr_model/'
    
        # Next build list of files to download
        cdcfiles = []
        for attr in attributes:
            cdcfiles.append('simple_'+attr+'.ckpt')
            cdcfiles.append('simple_'+attr+'_model.json')
            cdcfiles.append('simple_'+attr+'_weights.h5')
        
        files = []
        
        for f in cdcfiles:
            file = requests.get(os.path.join(cdcweb, f), allow_redirects=True)
            open(os.path.join(local_path, f), 'wb').write(file.content)
        
        # r=root, d=directories, f = files
        for r, d, f in os.walk(local_path):
            for attr in attributes:
                for file in f:
                    if attr in file:
                        files.append(os.path.join(r, file))
        
        print('Attribute files downloaded:')
        print(files)
        
        return files
        
    def dwnld_celebA_model(self, local_path):
        '''
        Download celebA model
        
        Args:
            local path (str): local path to where files are downloaded

        Returns:
            files (str list): list of files that were downloaded
        '''

        # This is the link where the celebA model is stored
        cdcweb = 'http://aix360-dev.mybluemix.net/static/CEM-MAF/celebA'
    
        files = []
        
        file = requests.get(cdcweb, allow_redirects=True)
        open(os.path.join(local_path, 'celebA'), 'wb').write(file.content)
        
        # r=root, d=directories, f = files
        for r, d, f in os.walk(local_path):
            for file in f:
                if 'celebA' == file:
                    files.append(os.path.join(r, file))
        
        print('celebA model file downloaded:')
        print(files)
 
        return files
 
    def dwnld_celebA_data(self, local_path, ids):
        '''
        Download celebA data files
        
        Args:
            local path (str): local path to where files are downloaded
            ids (int list): list of ids to download data for 
        
        Returns:
            files (str list): list of files that were downloaded
        '''

        # This is the link where celebA image data is stored
        cdcweb = 'http://aix360-dev.mybluemix.net/static/CEM-MAF/data/'
    
        # Next build list of files to download
        cdcfiles = []
        for id in ids:
            cdcfiles.append(str(id)+'_img.npy')
            cdcfiles.append(str(id)+'_latent.npy')
            cdcfiles.append(str(id)+'img.png')        
        files = []
        
        if not os.path.exists(local_path):
            os.makedirs(local_path)

        for f in cdcfiles:
            file = requests.get(os.path.join(cdcweb, f), allow_redirects=True)
            open(os.path.join(local_path, f), 'wb').write(file.content)
        
        # r=root, d=directories, f = files
        for r, d, f in os.walk(local_path):
            for id in ids:
                for file in f:
                    if str(id) in file:
                        files.append(os.path.join(r, file))
        
        print('Image files downloaded:')
        print(files)
        
        return files



import unittest

from sklearn.preprocessing import OneHotEncoder
import numpy as np

from aix360.algorithms.protodash import ProtodashExplainer
from aix360.algorithms.protodash import get_Gaussian_Data
from aix360.datasets.cdc_dataset import CDCDataset


class TestProtoDashExplainer(unittest.TestCase):

    def test_ProtoDashExplainer(self):

        # load NHANES CDC Questionnaire data
        nhanes = CDCDataset()

        # load income questionnaire dataset
        df = nhanes.get_csv_file('INQ_H.csv')
        print(df.head(5))

        data = df.to_numpy()
        np.set_printoptions(suppress=True)
        print(data[:5])

        # One-hot encode the data
        original = data

        # replace nan's with 0's
        original[np.isnan(original)] = 0

        # delete 1st column (contains sequence numbers)
        original = original[:, 1:]

        # one hot encoding of all columns/features
        onehot_encoder = OneHotEncoder(sparse=False)
        onehot_encoded = onehot_encoder.fit_transform(original)

        print(onehot_encoded.shape, original.shape)

        # Obtain an explanation for data
        Y = onehot_encoded
        X = Y
        explainer = ProtodashExplainer()
        (W, S, _) = explainer.explain(X, Y, m=5)

        dfs = df.iloc[S].copy()
        dfs["Weight"] = W
        print(dfs)

        # Gaussian (simulated) data example
        # generate normalized gaussian data X, Y with 100 features and 300 & 4000 observations respectively

        (X, Y) = get_Gaussian_Data(100, 300, 4000)
        print(X.shape, Y.shape)
        (W, S, _) = explainer.explain(X, Y, m=5, kernelType='Gaussian', sigma=2)

        print(S, W)

if __name__ == '__main__':
    unittest.main()




import unittest
import sys
import numpy as np
import os

sys.path.append(os.getcwd())
from aix360.algorithms.profwt import prof_weight_compute


class Testprofwexplainer(unittest.TestCase):

    def test_prof_weight_compute(self):


        """ Test Prof Weight Computation """

        ## Create a fixed probability matrix (samples=3,num_classes=9) and rotate it 3 times along the columns to generate 3 layers.
        ## The Labels are appropriately generated such that the prof_weight computation would compute the normalized sum of the first 3 probabilities for the 
        ## first sample, the next 3 probabilities for the second sample, the next 3 probabilities for the third sample etc..

        ## Generate a random probability vector and replicate it 3 times.
        a=np.random.random_sample((1,9))
        a=a/np.sum(a) 
        data_matrix=np.outer(np.ones((3,1)),a)
        np.save("./test_case_data_1",data_matrix)
        np.save("./test_case_data_2",np.roll(data_matrix,-1,axis=1))
        np.save("./test_case_data_3",np.roll(data_matrix,-2,axis=1))


        y=np.zeros((1,9))
        y[0,0]=1
        y1=np.roll(y,3,axis=1)
        y2=np.roll(y1,3,axis=1)
        Y=np.vstack((np.vstack((y,y1)),y2))

        np.save("./test_case_train_labels",Y)



        # Load the probes already stored numbered by layers.
        list_probe_filenames=['./test_case_data_'+str(x)+'.npy' for x in range(1,4)]
        print(list_probe_filenames)
        # Load the label corresponding to this
        train_label_path='./test_case_train_labels.npy'
        y_train=np.load(train_label_path)

        #Compute Prof-Weight by calling prof_weight_compute function
        start_layer=0
        final_layer=2
        w=prof_weight_compute(list_probe_filenames,start_layer,final_layer,y_train)
        w=w.reshape(w.shape[0],)
        self.assertTrue( (abs(3*w[0]-np.sum(a[0,0:3]))<0.0001) & (abs(3*w[1]-np.sum(a[0,3:6]))<0.0001) & (abs(3*w[2]-np.sum(a[0,6:9]))<0.0001), "Weight Computation has an error")

        print("prof_weight_compute Test passed..")
        #Delete the npy files created
        print("Removing Created test files....")
        if os.path.isfile(train_label_path):
            os.remove(train_label_path)
        for f in list_probe_filenames:
            if os.path.isfile(f):
                os.remove(f)



if __name__ == '__main__':
    unittest.main()



# Module for local white box explainer base classes for text, image, and tabular data. 
# All local WB explainer algorithms (e.g. CEM, etc.) would inherit these classes. 

import abc 
import sys

# Ensure compatibility with Python 2/3 
if sys.version_info >= (3, 4):
    ABC = abc.ABC
else:
    ABC = abc.ABCMeta(str('ABC'), (), {})


class LocalWBExplainer(ABC):
        
    """ 
    LocalWBExplainer is the base class for local post-hoc white box explainers (LBBE).
    Such explainers generally require access to model's internals beyond its predict function.
    Examples include Contrastive explanation method[#1]_, Layer-wise Relevance Propagation[#2]_, etc.

    References:
        .. [#] Explanations based on the Missing: Towards Contrastive Explanations with Pertinent Negatives,
           NIPS, 2018. Amit Dhurandhar, Pin-Yu Chen, Ronny Luss, Chun-Chen Tu,
           Paishun Ting, Karthikeyan Shanmugam, Payel Das. https://arxiv.org/abs/1802.07623
        .. [#2] http://www.heatmapping.org/

    """ 
    def __init__(self, *argv, **kwargs):
        
        """
        Constructor method, initialize a LocalBBExplainer object.
        ToDo: check common steps that need to be distilled here. 
        """
                
    @abc.abstractmethod 
    def set_params(self, *argv, **kwargs): 
        """
        Set parameters for the explainer. 
        """
        raise NotImplementedError 
        
    
    @abc.abstractmethod 
    def explain_instance(self, *argv, **kwargs):
        """
        Explain an input instance x.
        """
        raise NotImplementedError
        





import os,sys

import pandas as pd


class TEDDataset():
    """
    The goal of this synthetic dataset is to predict employee attrition from a
    fictious company.  The dataset is generated by a python file called
    ``GenerateData.py`` in the ``aix360/data/ted_data/`` directory.

    Like most datasets, each instance consists of a feature vector and a Y
    label, which represents whether the employee associated with the feature
    vector will leave the company.  However, unlike most datasets, each instance
    will also have an Explanation (E). This is motivated by the TED framework,
    which requires explanations in its training data, but can be used by other
    explainability algorithms as a metric for explainability.

    See also:
        * AIES'19 paper by Hind et al. [#]_ for more information on the TED
          framework.
        * The tutorial notebook ``TED_Cartesian_test.ipynb`` for information
          about how to use this dataset and the TED framework.
        * ``GenerateData.py`` for more information on how the dataset is
          generated or to created a tailored version of the dataset.

    References:
        .. [#] `Michael Hind, Dennis Wei, Murray Campbell, Noel C. F. Codella,
           Amit Dhurandhar, Aleksandra Mojsilovic, Karthikeyan Natesan Ramamurthy,
           Kush R. Varshney, "TED: Teaching AI to Explain its Decisions,"
           AAAI /ACM Conference on Artificial Intelligence, Ethics,
           and Society (AIES-19), 2019.
           <http://www.aies-conference.com/wp-content/papers/main/AIES-19_paper_128.pdf>`_
    """

    def __init__(self, dirpath=None):
        self._dirpath = dirpath
        if not self._dirpath:
            self._dirpath = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                '..', 'data','ted_data')


    def load_file(self, fileName='Retention.csv') :
        """ Open dataset file and populate X, Y, and E

        Args:
            fileName (String) : filename of dataset, a structured (CSV) dataset where

                - The first N-2 columns are the features (X).
                - The next to last column is the label (Y) {0, 1}
                - The last column gives the explanations (E) {0, 1, ..., MaxE}.  We assume the explanation space
                  is dense, i.e., if there are MaxE+1 unique explanations, they will be given IDs from 0 .. MaxE
                - first row contains header information for each column and should be "Y" for labels and "E" for explanations
                - each row is an instance
        Returns:
            tuple:
                * **X** -- list of features vectors
                * **Y** -- list of labels
                * **E** -- list of explanations
        """
        self._filepath = os.path.join(self._dirpath, fileName)

        try:
            data = pd.read_csv(self._filepath)
            X = data.iloc[:,:-2]   # Choose all rows and all cols, except for the last 2 cols
            Y = data['Y']          # Choose col with header 'Y'
            E = data['E']          # Choose col with header 'E'

            return X, Y, E

        except IOError as err:
            print("IOError: {}".format(err))
            sys.exit(1)



import abc
import sys
import numpy as np


# compatibility with Python 2 and 3 when using ABCMeta
if sys.version_info >= (3, 4):
    ABC = abc.ABC
else:
    ABC = abc.ABCMeta(str('ABC'), (), {})


class BaseClassifier(ABC):

    """
    Base class for classifiers (image, tabular data, etc.)
    The base class is meant to be framework independent so that each
    framework specific classes (keras, torch, etc.) could inherit and
    implement these methods.
    """
    def __init__(self):

        """
        Initialize a Classifier object.
        """

    @abc.abstractmethod
    def predict(self, x, verbose=0):
        """
        Prediction on a batch of inputs x.
        type: numpy.ndarray
        """
        raise NotImplementedError


    @property
    def nb_classes(self):
        """
        Returns the number of output classes
        """
        return self._nb_classes

    @property
    def input_shape(self):
        """
        Returns the shape of one input to the classifier
        """
        return self._input_shape


class KerasClassifier(BaseClassifier):
    """Keras classifier has the following attributes:

    Attributes:
        _model: the compiled keras classifier
        _input:  input placeholder of the model
        _output: output placeholder of the model
        _nb_classes:  number of outputs (e.g. binary classifier has 1 output)
        _input_shape: shape of 1 input sample
    """
    def __init__(self, model, input_layer=0, output_layer=0):

        """Initialize KerasClassifier.

        Args:
            model: a trained keras classifier model.
        """

        import keras.backend as k

        super(KerasClassifier, self).__init__()

        self._model = model

        if hasattr(model, 'inputs'):
            self._input = model.inputs[input_layer]
        else:
            self._input = model.input

        if hasattr(model, 'outputs'):
            self._output = model.outputs[output_layer]
        else:
            self._output = model.output

        _, self._nb_classes = k.int_shape(self._output)
        self._input_shape   = k.int_shape(self._input)[1:]

    def predict(self, x, verbose=0):
        """
        Make predictions on batch of vector inputs
        """
        return(self._model.predict(x, verbose=verbose))

    def predict_classes(self, x, verbose=0):
        """
        returns classes instead of probabilities
        """
        return(self._model.predict_classes(x, verbose=verbose))

    def predict_long(self, x):
        prob = self.predict(x)
        predicted_class = np.argmax(prob)
        prob_str = np.array2string(prob).replace('\n','')
        return(prob, predicted_class, prob_str)

    def predictsym(self, x):
        return self._model(x)



import os

import pandas as pd
import numpy as np
from sklearn.preprocessing import LabelEncoder
from sklearn.model_selection import train_test_split
from keras.utils import to_categorical

def nan_preprocessing(df):
    # Separate and encode target variable as 0,1
    y = df.pop('RiskPerformance')
    enc = LabelEncoder()
    y = enc.fit_transform(y)
    # Remove rows where all data is missing
    y = y[(df > -9).any(axis=1)]
    df = df.loc[(df > -9).any(axis=1)]
    # Replace remaining special values with NaN
    df = df[df >= 0]
    df['RiskPerformance'] = y
    return df

def default_preprocessing(df):
    # Details and preprocessing for FICO dataset
    x_cols, y_col = df.columns[0:-1], df.columns[-1]

    # Preprocessing the HELOC dataset
    # Remove all the rows containing -9 in the ExternalRiskEstimate column
    # df = df[df.ExternalRiskEstimate != -9]
    # add columns for -7 and -8 in the dataset
    for col in x_cols:
        df[col][df[col].isin([-7, -8, -9])] = 0
    # Get the column names for the covariates and the dependent variable
    df = df[(df[x_cols].T != 0).any()]
    x = df.values[:, 0:-1]

    # encode target variable ('bad', 'good')
    cat_values = df[y_col].values
    enc = LabelEncoder()
    enc.fit(cat_values)
    num_values = enc.transform(cat_values)
    y = np.array(num_values)

    return np.hstack((x, y.reshape(y.shape[0], 1)))


class HELOCDataset():
    """HELOC Dataset.

    The FICO HELOC dataset [#]_ contains anonymized information about home equity line of credit (HELOC) applications 
    made by real homeowners. A HELOC is a line of credit typically offered by a US bank as a percentage of home
    equity (the difference between the current market value of a home and the outstanding balance of all liens,
    e.g. mortgages). The customers in this dataset have requested a credit line in the range of USD 5,000 - 150,000.

    The target variable in this dataset is a binary variable called RiskPerformance. The value “Bad” indicates that an
    applicant was 90 days past due or worse at least once over a period of 24 months from when the credit account
    was opened. The value “Good” indicates that they have made their payments without ever being more than 90 days
    overdue.

    This dataset can be used to train a machine learning model to predict whether the homeowner qualifies for a line
    of credit or not. The HELOC dataset and more information about it, including instructions to download are available in
    the reference below.

    References:
        .. [#] `Explainable Machine Learning Challenge - FICO Community.
           <https://community.fico.com/s/explainable-machine-learning-challenge?tabset-3158a=2>`_

    """

    def __init__(self, custom_preprocessing=default_preprocessing, dirpath=None):
        self._dirpath = dirpath
        if not self._dirpath:
            self._dirpath = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                '..', 'data','heloc_data')

        self._filepath = os.path.join(self._dirpath, 'heloc_dataset.csv')
        try:
            #require access to dataframe
            #df = pd.read_csv(filepath)
            self._df = pd.read_csv(self._filepath)
        except IOError as err:
            print("IOError: {}".format(err))
            print("To use this class, please place the heloc_dataset.csv:")
            print("file, as-is, in the folder:")
            print("\n\t{}\n".format(os.path.abspath(os.path.join(
               os.path.abspath(__file__), '..', 'data','heloc_data'))))
            import sys
            sys.exit(1)

        if custom_preprocessing:
            #require access to dataframe
            #self._data = custom_preprocessing(df)
            self._data = custom_preprocessing(self._df.copy())

    #require access to the dataframe
    def dataframe(self):
        return(self._df)

    def data(self):
        return self._data

    def split(self, random_state=0):
        (data_train, data_test) = train_test_split(self._data, stratify=self._data[:,-1], random_state=random_state)

        x_train = data_train[:,0:-1]
        x_test = data_test[:, 0:-1]
        y_train = data_train[:, -1]
        y_test = data_test[:, -1]

        y_train_b = to_categorical(y_train)
        y_test_b = to_categorical(y_test)

        return (self._data, x_train, x_test, y_train_b, y_test_b)



## aen_attack.py -- attack a network optimizing elastic-net distance with an en decision rule
##                  when autoencoder loss is applied
##
## Copyright (C) 2018, PaiShun Ting <paishun@umich.edu>
##                     Chun-Chen Tu <timtu@umich.edu>
##                     Pin-Yu Chen <Pin-Yu.Chen@ibm.com>
## Copyright (C) 2017, Yash Sharma <ysharma1126@gmail.com>.
## Copyright (C) 2016, Nicholas Carlini <nicholas@carlini.com>.
##
## This program is licenced under the BSD 2-Clause licence,
## contained in the LICENCE file in this directory.

import sys
import tensorflow as tf
import numpy as np
import pickle
from tensorflow.contrib.keras.api.keras.models import Model, Sequential, model_from_json
from tensorflow.contrib.keras.api.keras.callbacks import ModelCheckpoint
import os

class AEADEN:
    def __init__(self, sess, model, attributes, aix360_path, mode, batch_size, kappa, init_learning_rate, binary_search_steps, max_iterations, initial_const, gamma, attr_reg, attr_penalty_reg, latent_square_loss_reg):
        """
        Initialize PN explainer object. 
        
        Args:
            sess (tensorflow.python.client.session.Session): Tensorflow session
            model: KerasClassifier that contains a trained model to be explained
            attributes (str list): list of attributes to load attribute classifiers for
            aix360_path (str): path to aix360 used to determine paths to pretrained attribute classifiers  
            mode (str): "PN" for pertinent negative or "PP" for pertinent positive
            batch_size (int): batch size for how many instances to explain
            kappa (float): Confidence parameter that controls difference between prediction of
                PN (or PP) and original prediction
            init_learning_rate (float): initial learning rate for gradient descent optimizer
            binary_search_steps (int): Controls number of random restarts to find best PN
            max_iterations (int): Max number iterations to run some version of gradient descent on
                PN optimization problem from a single random initialization, i.e., total 
                number of iterations wll be arg_binary_search_steps * arg_max_iterations
            initial_const (int): Constant used for upper/lower bounds in binary search
            gamma (float): Penalty parameter encouraging addition of attributes for PN
            attr_reg (float): Penalty parameter on regularization of PN to be predicted different from 
                original image
            attr_penalty_reg (float): Penalty regularizing PN from being too different from original image
            latent_square_loss_reg (float): Penalty regularizing PN from being too different from original 
                image in the latent space
        """


#        image_size, num_channels, nun_classes = model.image_size, model.num_channels, model.num_labels
        # %%change%%
        image_size = model._input_shape[0]
        num_channels = model._input_shape[2]
        nun_classes = model._nb_classes 
        shape = (batch_size, image_size, image_size, num_channels)
        latent_shape = (batch_size, 512)

        self.sess = sess
        self.INIT_LEARNING_RATE = init_learning_rate
        self.MAX_ITERATIONS = max_iterations
        self.BINARY_SEARCH_STEPS = binary_search_steps
        self.kappa = kappa
        self.init_const = initial_const
        self.batch_size = batch_size
        self.mode = mode
        self.gamma = gamma
        self.attributes = attributes
        self.aix360_path = aix360_path
        self.attr_reg = attr_reg
        self.attr_penalty_reg = attr_penalty_reg
        self.attr_threshold = tf.constant(0.5, dtype="float32") # penalize the attributes of orig_img having scores <= this value
        self.latent_square_loss_reg = latent_square_loss_reg

        # these are variables to be more efficient in sending data to tf
        self.orig_img = tf.Variable(np.zeros(shape), dtype=tf.float32)
        self.orig_latent = tf.Variable(np.zeros(latent_shape), dtype=tf.float32, name="orig_latent")
        self.adv_latent = tf.Variable(np.zeros(latent_shape), dtype=tf.float32, name="adv_latent")
        self.target_lab = tf.Variable(np.zeros((batch_size,nun_classes)), dtype=tf.float32)
        self.const = tf.Variable(np.zeros(batch_size), dtype=tf.float32)
        self.global_step = tf.Variable(0.0, trainable=False)

        # and here's what we use to assign them
        self.assign_orig_img = tf.placeholder(tf.float32, shape)
        self.assign_orig_latent = tf.placeholder(tf.float32, latent_shape, name="assign_orig_latent")
        self.assign_adv_latent = tf.placeholder(tf.float32, latent_shape, name="assign_adv_latent")
        self.assign_target_lab = tf.placeholder(tf.float32, (batch_size,nun_classes), name="assign_target_label")
        self.assign_const = tf.placeholder(tf.float32, [batch_size])

        ### Load attribute classifier
        nn_type = "simple"
        #import copy
        attr_model_list=[]
 #       attr_threshold_idx=[]
 #       count=0
        for attr in self.attributes:
            # load test data into memory using Image Data Generator
#            print("Loading data for " + attr + " into memory")
            # load json and create model
            json_file_name = os.path.join(aix360_path, "models/CEM_MAF/{}_{}_model.json".format(nn_type, attr))
            json_file = open(json_file_name, 'r')
            loaded_model_json = json_file.read()
            json_file.close()
            loaded_model = model_from_json(loaded_model_json)
            # load weights into new model
            weight_file_name = os.path.join(aix360_path, "models/CEM_MAF/{}_{}_weights.h5".format(nn_type, attr))
            loaded_model.load_weights(weight_file_name)
            print("Loaded model for " + attr + " from disk")
            attr_model_list.append(loaded_model)
#            if self.mode == "PP":
#                pass
#            else:
#                attr_threshold_idx=tf.cond(loaded_model(self.orig_img)<= self.attr_threshold, lambda: attr_threshold_idx.append(count), lambda: attr_threshold_idx)
#            count=count+1
        print("# of attr models is",len(attr_model_list))
#        print("# of attr smaller than THR is",len(attr_threshold_idx))

        # load gan
        # Import official CelebA-HQ networks.
        with open(os.path.join(aix360_path, 'algorithms/contrastive/progressive_growing_of_gans/karras2018iclr-celebahq-1024x1024.pkl'), 'rb') as file:
            G, D, Gs = pickle.load(file)
        # inputs, redundent ...
        # in_labels = tf.placeholder(tf.float32, shape=(None, 0))
        in_labels = tf.constant(0, shape=(1, 0))
        # get the GAN network, 
        with tf.variable_scope(Gs.scope, reuse=tf.AUTO_REUSE):
            out_image = Gs._build_func(self.adv_latent , in_labels, **Gs.static_kwargs, reuse=True)
            tanspose_image = tf.transpose(out_image, perm=[0, 2, 3, 1])
            resize_image= tf.image.resize_images(tanspose_image, [224, 224])
            self.adv_img = tf.clip_by_value(resize_image/2, -0.5, 0.5)
        
        self.adv_updater = tf.assign(self.adv_latent, self.assign_adv_latent)

        """--------------------------------"""
        # prediction BEFORE-SOFTMAX of the model
        self.delta_img = self.orig_img - self.adv_img
        # %%change%%
        if self.mode == "PP":
#            self.ImgToEnforceLabel_Score = model.predict(self.delta_img)
            self.ImgToEnforceLabel_Score = model.predictsym(self.delta_img)
        elif self.mode == "PN":
#            self.ImgToEnforceLabel_Score = model.predict(self.adv_img)
            self.ImgToEnforceLabel_Score = model.predictsym(self.adv_img)
        # Attribute classifier score
        self.attr_score = tf.constant(0, dtype="float32")        
        self.attr_penalty = tf.constant(0, dtype="float32")  
        if self.mode == "PP":
            for i in range(len(attr_model_list)):
                self.attr_score = self.attr_score + tf.maximum(attr_model_list[i](self.adv_img) - attr_model_list[i](self.orig_img),tf.constant(0, tf.float32))
                self.attr_score = tf.squeeze(self.attr_score) 
           #self.ImgToEnforceLabel_Score = model.predict(self.adv_img)
            #self.ImgToEnforceLabel_Score_s = model.predict(self.adv_img_s)
        elif self.mode == "PN":
            for i in range(len(attr_model_list)):
                self.attr_score = self.attr_score + tf.maximum(attr_model_list[i](self.orig_img) - attr_model_list[i](self.adv_img),tf.constant(0, tf.float32))
                self.attr_score = tf.squeeze(self.attr_score)        
                #self.attr_penalty = tf.squeeze(self.attr_penalty)                 
                self.attr_penalty = self.attr_penalty + tf.multiply(tf.cond(tf.squeeze(attr_model_list[i](self.orig_img)) <= self.attr_threshold, lambda: tf.constant(1, tf.float32), lambda: tf.constant(0, tf.float32)),tf.squeeze(attr_model_list[i](self.adv_img)))
                #tf.maximum(attr_model_list[i](self.orig_img) - attr_model_list[i](self.adv_img),tf.constant(0, tf.float32))
        # Sum of attributes penalty in attr_threshold_idx  
 #       self.attr_penalty = tf.constant(0, dtype="float32")
 #       if len(attr_threshold_idx)==0:
 #           pass
 #       else:  
 #           for i in range(len(attr_threshold_idx)):
 #                self.attr_penalty =  self.attr_penalty + attr_model_list[i](self.adv_img)

        self.delta_latent = self.orig_latent - self.adv_latent

        # distance to the input data
        self.L2_img_dist = tf.reduce_sum(tf.square(self.delta_img),[1,2,3])
        self.L2_latent_dist = tf.reduce_sum(tf.square(self.delta_latent))



        # compute the probability of the label class versus the maximum other
        self.target_lab_score        = tf.reduce_sum((self.target_lab)*self.ImgToEnforceLabel_Score,1)
        self.max_nontarget_lab_score = tf.reduce_max((1-self.target_lab)*self.ImgToEnforceLabel_Score - (self.target_lab*10000),1)
        if self.mode == "PP":
            Loss_Attack = tf.maximum(0.0, self.max_nontarget_lab_score - self.target_lab_score + self.kappa)
        elif self.mode == "PN":
            Loss_Attack = tf.maximum(0.0, -self.max_nontarget_lab_score + self.target_lab_score + self.kappa)
        # sum up the losses
        self.Loss_Latent_L2Dist    = tf.reduce_sum(self.latent_square_loss_reg*self.L2_latent_dist)
        self.Loss_Img_L2Dist    = tf.reduce_sum(self.gamma*self.L2_img_dist)
        self.Loss_Attack    = tf.reduce_sum(self.const*Loss_Attack)
        self.Loss_attr = tf.reduce_sum(self.attr_reg*self.attr_score)
        self.Loss_attr_penalty = tf.reduce_sum(self.attr_penalty_reg*self.attr_penalty)
        self.Loss_Overall    = self.Loss_Latent_L2Dist + self.Loss_Img_L2Dist + self.Loss_Attack + self.Loss_attr + self.Loss_attr_penalty
        # self.Loss_Overall    = self.Loss_Attack

        self.learning_rate = tf.train.polynomial_decay(self.INIT_LEARNING_RATE, self.global_step, self.MAX_ITERATIONS, 0, power=0.5)
        optimizer = tf.train.GradientDescentOptimizer(self.learning_rate)
        start_vars = set(x.name for x in tf.global_variables())
        self.train = optimizer.minimize(self.Loss_Overall, var_list=[self.adv_latent], global_step=self.global_step)
        end_vars = tf.global_variables()
        new_vars = [x for x in end_vars if x.name not in start_vars]

        # these are the variables to initialize when we run
        self.setup = []
        self.setup.append(self.orig_img.assign(self.assign_orig_img))
        self.setup.append(self.orig_latent.assign(self.assign_orig_latent))
        self.setup.append(self.adv_latent.assign(self.assign_adv_latent))
        self.setup.append(self.target_lab.assign(self.assign_target_lab))
        self.setup.append(self.const.assign(self.assign_const))
        

        self.init = tf.variables_initializer(var_list=[self.global_step]+[self.adv_latent]+new_vars)

    def attack(self, imgs, labs, latent):
        """
        Find PN for an input instance input_image e.g. celebA is shape (1, 224, 224, 3)
        
        Input:
            imgs (numpy.ndarry): images to be explained, of shape (num_images, size, size, channels)
            labs: one hot encoded vectors of target labels for PN, i.e. which labels
                to explain with an image of a different class
            latent (numpy.ndarry): image to be explained, of shape (1, size, size, channels)
                in the latent space
                
        Output: 
            adv_img (numpy.ndarry): the pertinent negative image
        """
        def compare(x,y):
            if not isinstance(x, (float, int, np.int64)):
                x = np.copy(x)
                # x[y] -= self.kappa if self.PP else -self.kappa
                if self.mode == "PP":
                    x[y] -= self.kappa
                elif self.mode == "PN":
                    x[y] += self.kappa
                x = np.argmax(x)
            if self.mode == "PP":
                return x==y
            else: 
                return x!=y

        batch_size = self.batch_size

        # set the lower and upper bounds accordingly
        Const_LB = np.zeros(batch_size)
        CONST = np.ones(batch_size)*self.init_const
        Const_UB = np.ones(batch_size)*1e10
        # the best l2, score, and image attack
        overall_best_dist = [1e10]*batch_size
        overall_best_attack = [np.zeros(imgs[0].shape)]*batch_size

        for binary_search_steps_idx in range(self.BINARY_SEARCH_STEPS):
            # completely reset adam's internal state.
            self.sess.run(self.init)
            img_batch = imgs[:batch_size]
            label_batch = labs[:batch_size]

            current_step_best_dist = [1e10]*batch_size
            current_step_best_score = [-1]*batch_size

            # set the variables so that we don't have to send them over again
            self.sess.run(self.setup, {self.assign_orig_img: img_batch,
                                       self.assign_target_lab: label_batch,
                                       self.assign_const: CONST,
                                       self.assign_adv_latent: latent,
                                       self.assign_orig_latent: latent
                                       })

            for iteration in range(self.MAX_ITERATIONS):
                # perform the attack

                self.sess.run([self.train])
                temp_adv_latent = self.sess.run(self.adv_latent)
                self.sess.run(self.adv_updater, feed_dict={self.assign_adv_latent: temp_adv_latent})

                Loss_Overall, OutputScore, adv_img = self.sess.run([self.Loss_Overall, self.ImgToEnforceLabel_Score, self.adv_img])
                Loss_Attack, Loss_Latent_L2Dist, Loss_Img_L2Dist, Loss_attr = self.sess.run([self.Loss_Attack, self.Loss_Latent_L2Dist, self.Loss_Img_L2Dist, self.attr_score])
                target_lab_score, max_nontarget_lab_score_s = self.sess.run([self.target_lab_score, self.max_nontarget_lab_score])


                

                # if iteration%(self.MAX_ITERATIONS//10) == 0:
                if iteration % 10 == 0:
                    print("iter:{} const:{}". format(iteration, CONST))
                    print("Loss_Overall:{:.4f}, Loss_Attack:{:.4f}, Loss_attr:{:.4f}". format(Loss_Overall, Loss_Attack, Loss_attr))
                    print("Loss_Latent_L2Dist:{:.4f}, Loss_Img_L2Dist:{:.4f}". format(Loss_Latent_L2Dist, Loss_Img_L2Dist))
                    print("target_lab_score:{:.4f}, max_nontarget_lab_score:{:.4f}". format(target_lab_score[0], max_nontarget_lab_score_s[0]))
                    print("")
                    sys.stdout.flush()

                for batch_idx,(the_dist, the_score, the_adv_img) in enumerate(zip([Loss_Overall], OutputScore, adv_img)):
                    if the_dist < current_step_best_dist[batch_idx] and compare(the_score, np.argmax(label_batch[batch_idx])):
                        current_step_best_dist[batch_idx] = the_dist
                        current_step_best_score[batch_idx] = np.argmax(the_score)
                    if the_dist < overall_best_dist[batch_idx] and compare(the_score, np.argmax(label_batch[batch_idx])):
                        overall_best_dist[batch_idx] = the_dist
                        overall_best_attack[batch_idx] = the_adv_img

            # adjust the constant as needed
            for batch_idx in range(batch_size):
                if compare(current_step_best_score[batch_idx], np.argmax(label_batch[batch_idx])) and current_step_best_score[batch_idx] != -1:
                    # success, divide const by two
                    Const_UB[batch_idx] = min(Const_UB[batch_idx],CONST[batch_idx])
                    if Const_UB[batch_idx] < 1e9:
                        CONST[batch_idx] = (Const_LB[batch_idx] + Const_UB[batch_idx])/2
                else:
                    # failure, either multiply by 10 if no solution found yet
                    #          or do binary search with the known upper bound
                    Const_LB[batch_idx] = max(Const_LB[batch_idx],CONST[batch_idx])
                    if Const_UB[batch_idx] < 1e9:
                        CONST[batch_idx] = (Const_LB[batch_idx] + Const_UB[batch_idx])/2
                    else:
                        CONST[batch_idx] *= 10

        # return the best solution found
        overall_best_attack = overall_best_attack[0]
        return overall_best_attack.reshape((1,) + overall_best_attack.shape)



from ..dise import DISExplainer


class GLRMExplainer(DISExplainer):
    """
    Generalized Linear Rule Model explainer. Provides access to the following
    directly interpretable supervised learning methods:

    * **Linear Rule Regression:** linear regression on rule-based features [#Icml2019]_.
    * **Logistic Rule Regression:** logistic regression on rule-based features [#Icml2019]_.

    References:
        .. [#Icml2019] `D. Wei, S. Dash, T. Gao, O. Günlük, "Generalized linear rule
           models." International Conference on Machine Learning (ICML), 2019.
           <http://proceedings.mlr.press/v97/wei19a/wei19a.pdf>`_
    """

    def __init__(self, model):
        """
        Initialize a GLRMExplainer object.

        Args:
          model: model to operate on. Instance of either

            * :class:`aix360.algorithms.rbm.linear_regression.LinearRuleRegression` or
            * :class:`aix360.algorithms.rbm.logistic_regression.LogisticRuleRegression`
        """
        super(GLRMExplainer, self).__init__()
        self._model = model

    def set_params(self, *argv, **kwargs):
        """
        Set parameters for the explainer.
        """
        pass

    def fit(self, X_train, Y_train, Xstd=None):
        """Fit model to training data.

        Args:
            X_train (DataFrame): Binarized features with MultiIndex column labels
            Y_train (array): Target variable
            Xstd (DataFrame, optional): Standardized numerical features
        Returns:
            GLRMExplainer: Self
        """
        self._model.fit(X_train, Y_train, Xstd)

    def predict(self, X, Xstd=None):
        """Predict responses.

        Args:
            X (DataFrame): Binarized features with MultiIndex column labels
            Xstd (DataFrame, optional): Standardized numerical features
        Returns:
            array: y -- Predicted responses
        """
        return self._model.predict(X, Xstd)

    def predict_proba(self, X, Xstd=None):
        """Predict probabilities of Y=1. Only available if underlying model implements
        `predict_proba` method.

        Args:
            X (DataFrame): Binarized features with MultiIndex column labels
            Xstd (DataFrame, optional): Standardized numerical features
        Returns:
            array: p -- Predicted probabilities

        Raises:
            ValueError: if model doesn't implement `predict_proba`
        """
        if hasattr(self._model, 'predict_proba'):
            return self._model.predict_proba(X, Xstd)
        raise ValueError("Current model doesn't predict probabilities")

    def explain(self, maxCoeffs=None, highDegOnly=False, prec=2):
        """Return DataFrame holding model features and their coefficients.

        Args:
            maxCoeffs (int, optional): Maximum number of rules/numerical features to show
            highDegOnly (bool, optional): Only show higher-degree rules
            prec (int, optional): Number of decimal places to show for floating-value thresholds
        Returns:
            DataFrame: dfExpl -- Rules/numerical features and their coefficients
        """
        return self._model.explain(maxCoeffs, highDegOnly, prec)

    def visualize(self, Xorig, fb, features=None):
        """Plot generalized additive model component, which includes first-degree rules 
        and linear functions of unbinarized ordinal features but excludes higher-degree rules.

        Args:
            Xorig (DataFrame): Original unbinarized features
            fb: FeatureBinarizer object used to binarize features
            features (list, optional): Subset of features to be plotted
        """
        return self._model.visualize(Xorig, fb, features)



import numpy as np
import json
import sys,os
import urllib.request
import tarfile
import pickle as cp
from sklearn.preprocessing import OneHotEncoder
import shutil

class CIFARDataset():
    """
    The CIFAR-10 dataset [#]_ consists of 60000 32x32 color images. Target variable is one amongst 10 classes. The dataset has
    6000 images per class. There are 50000 training images and 10000 test images. The classes are: airplane, automobile,
    bird, cat, deer, dog, frog, horse, ship ,truck. We further divide the training set into train1 (30000 samples) and
    train2 (20,000 samples). For ProfWt, the complex model is trained on train1 while the simple model is trained on train2.

    References:
        .. [#] `Krizhevsky, Hinton. Learning multiple layers of features from tiny images. Technical Report, University of
           Toronto 1 (4), 7. 2009 <https://www.cs.toronto.edu/~kriz/cifar.html>`_
    """

    def __init__(self, dirpath=None):
        self._dirpath = dirpath
        if not self._dirpath:
            self._dirpath = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                '..', 'data','cifar_data')
        self._download_data()

    def _download_data(self):
        name = 'cifar-10-python.tar.gz'
        json_file_name = 'cifar-10-train1-image.json'
        full_name = os.path.join(self._dirpath, name)
        if not os.path.exists(os.path.join(self._dirpath, json_file_name)):
            if not os.path.exists(full_name):
                print("retrieving file", name)
                urllib.request.urlretrieve('https://www.cs.toronto.edu/~kriz/' + name, full_name)
                print("retrieved")
            #now extract the files
            #print("extracting files")
            tar = tarfile.open(full_name, "r:gz")
            tar.extractall(self._dirpath)
            tar.close()        
            #print("extracted files")
            
            self._process_data()

            #now cleanup
            if os.path.exists(full_name):
                os.remove(full_name)
            
    def _process_data(self):
        image_size=32
        num_classes=10
        per_file_size=10000
        
        print("processing files...")
        datafile_path = os.path.join(self._dirpath, 'cifar-10-batches-py')
        for i in range(5):
            with open(os.path.join(datafile_path,'data_batch_'+str(i+1)), 'rb') as fileobj:
                dictionary = cp.load(fileobj, encoding='bytes')
                dum=dictionary[b'data'].reshape((per_file_size,3,32,32))
                dum_1=np.transpose(dum,(0,2,3,1)).astype('uint8')
                lab=np.asarray(dictionary[b'labels']).reshape((per_file_size,1))
                if i==0:
                    x_train=dum_1
                    y_train=lab     
                else:
                    x_train=np.concatenate((x_train,dum_1),0)
                    y_train=np.concatenate((y_train,lab),0)
        
        y_train=OneHotEncoder(sparse=False).fit_transform(y_train).astype('uint8')    
                
        assert x_train.shape==(5*per_file_size,image_size,image_size,3)
        assert y_train.shape==(5*per_file_size,num_classes)
        
        x_train.astype(float)/255
        
        x_train_1=x_train[0:30000,:,:,:]
        x_train_2=x_train[30000:,:,:,:]
        y_train_1=y_train[0:30000,:]
        y_train_2=y_train[30000:,:]
        
        
        with open(os.path.join(datafile_path,'test_batch'), 'rb') as fileobj:
            dictionary = cp.load(fileobj, encoding='bytes')
            x_test=dictionary[b'data'].reshape((per_file_size,3,32,32))
            x_test=np.transpose(x_test,[0,2,3,1]).astype('uint8')
            y_test=np.asarray(dictionary[b'labels']).reshape((per_file_size,1))
            y_test=OneHotEncoder(sparse=False).fit_transform(y_test).astype('uint8')
        
        with open(os.path.join(self._dirpath,'cifar-10-train1-image.json'),'w') as outfile:
            print("writing ",os.path.join(self._dirpath,'cifar-10-train1-image.json'))
            json.dump(x_train_1.tolist(),outfile)
        outfile.close()
        
        with open(os.path.join(self._dirpath,'./cifar-10-train2-image.json'),'w') as outfile:
            print("writing ",os.path.join(self._dirpath,'cifar-10-train2-image.json'))
            json.dump(x_train_2.tolist(),outfile)
        outfile.close()
        
        with open(os.path.join(self._dirpath,'./cifar-10-test-image.json'),'w') as outfile:
            print("writing ",os.path.join(self._dirpath,'cifar-10-test-image.json'))
            json.dump(x_test.tolist(),outfile)
        outfile.close()
        
        with open(os.path.join(self._dirpath,'./cifar-10-train1-label.json'),'w') as outfile:
            print("writing ",os.path.join(self._dirpath,'cifar-10-train1-label.json'))
            json.dump(y_train_1.tolist(),outfile)
        outfile.close()
        
        with open(os.path.join(self._dirpath,'./cifar-10-train2-label.json'),'w') as outfile:
            print("writing ",os.path.join(self._dirpath,'cifar-10-train2-label.json'))
            json.dump(y_train_2.tolist(),outfile)
        outfile.close()
        
        with open(os.path.join(self._dirpath,'./cifar-10-test-label.json'),'w') as outfile:
            print("writing ",os.path.join(self._dirpath,'cifar-10-test-label.json'))
            json.dump(y_test.tolist(),outfile)
        outfile.close()
        print("processing completed")
        #cleanup
        if os.path.exists(datafile_path):
            shutil.rmtree(datafile_path)
        

        
                        
    def load_file(self, filename):
        try:
            with open(os.path.join(self._dirpath, filename)) as file:
                data=json.load(file)
            file.close()
        except IOError as err:
            print("IOError: {}".format(err))
            sys.exit(1)
        return np.array(data)



import numpy as np
#import torch.nn as nn
import torch
import os
#import torch.nn.functional as F
import torchvision.datasets as dset
import torchvision.transforms as transforms
#from torch.autograd import Variable
#from torch.nn import Linear



class FMnistDataset():
    """
    Fashion-MNIST [#]_ is a large-scale image dataset of various fashion items (T-shirt/top, Trouser,
    Pullover, Dress. Coat, Sandal, Shirt, Sneaker, Bag, Ankle boot. The images are grayscale and 28x28
    in size with each image belong to one the above mentioned 10 categories. The training set contains
    60000 examples and the test set contains 10000 examples.

    References:
        .. [#] `Xiao, Han, Kashif Rasul, and Roland Vollgraf. Fashion-mnist: a
           novel image dataset for benchmarking machine learning algorithms.
           <https://arxiv.org/abs/1708.07747>`_
    """
    def __init__(self, batch_size=256, subset_size=50000, test_batch_size=256, dirpath=None):
        trans = transforms.Compose([transforms.ToTensor()])

        self._dirpath = dirpath
        if not self._dirpath:
            self._dirpath = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                '..', 'data','fmnit_data')


        train_set = dset.FashionMNIST(root=self._dirpath, train=True, transform=trans, download=True)
        test_set = dset.FashionMNIST(root=self._dirpath, train=False, transform=trans, download=True)

        indices = torch.randperm(len(train_set))[:subset_size]
        train_set = torch.utils.data.Subset(train_set, indices)

        self.train_loader = torch.utils.data.DataLoader(
            dataset=train_set,
            batch_size=batch_size,
            shuffle=True)
        self.test_loader = torch.utils.data.DataLoader(
            dataset=test_set,
            batch_size=test_batch_size,
            shuffle=False)

        self.name = "fmnist"
        self.data_dims = [28, 28, 1]
        self.train_size = len(self.train_loader)
        self.test_size = len(self.test_loader)
        self.range = [0.0, 1.0]
        self.batch_size = batch_size
        self.num_training_instances = len(train_set)
        self.num_test_instances = len(test_set)
        self.likelihood_type = 'gaussian'
        self.output_activation_type = 'sigmoid'

    def next_batch(self):
        for x, y in self.train_loader:
            x = np.reshape(x, (-1, 28, 28, 1))
            yield x, y

    def next_test_batch(self):
        for x, y in self.test_loader:
            x = np.reshape(x, (-1, 28, 28, 1))
            yield x, y



import numpy as np

def faithfulness_metric(model, x, coefs, base):
    """ This metric evaluates the correlation between the importance assigned by the interpretability algorithm
    to attributes and the effect of each of the attributes on the performance of the predictive model.
    The higher the importance, the higher should be the effect, and vice versa, The metric evaluates this by
    incrementally removing each of the attributes deemed important by the interpretability metric, and
    evaluating the effect on the performance, and then calculating the correlation between the weights (importance)
    of the attributes and corresponding model performance. [#]_

    References:
        .. [#] `David Alvarez Melis and Tommi Jaakkola. Towards robust interpretability with self-explaining
           neural networks. In S. Bengio, H. Wallach, H. Larochelle, K. Grauman, N. Cesa-Bianchi, and R. Garnett, editors,
           Advances in Neural Information Processing Systems 31, pages 7775-7784. 2018.
           <https://papers.nips.cc/paper/8003-towards-robust-interpretability-with-self-explaining-neural-networks.pdf>`_

    Args:
        model: Trained classifier, such as a ScikitClassifier that implements
            a predict() and a predict_proba() methods.
        x (numpy.ndarray): row of data.
        coefs (numpy.ndarray): coefficients (weights) corresponding to attribute importance.
        base ((numpy.ndarray): base (default) values of attributes

    Returns:
        float: correlation between attribute importance weights and corresponding effect on classifier.
    """

    #find predicted class
    pred_class = model.predict(x.reshape(1,-1))[0]

    #find indexs of coefficients in decreasing order of value
    ar = np.argsort(-coefs)  #argsort returns indexes of values sorted in increasing order; so do it for negated array
    pred_probs = np.zeros(x.shape[0])
    for ind in np.nditer(ar):
        x_copy = x.copy()
        x_copy[ind] = base[ind]
        x_copy_pr = model.predict_proba(x_copy.reshape(1,-1))
        pred_probs[ind] = x_copy_pr[0][pred_class]

    return -np.corrcoef(coefs, pred_probs)[0,1]


def monotonicity_metric(model, x, coefs, base):

    """ This metric measures the effect of individual features on model performance by evaluating the effect on
    model performance of incrementally adding each attribute in order of increasing importance. As each feature
    is added, the performance of the model should correspondingly increase, thereby resulting in monotonically
    increasing model performance. [#]_

    References:
        .. [#] `Ronny Luss, Pin-Yu Chen, Amit Dhurandhar, Prasanna Sattigeri, Karthikeyan Shanmugam, and
           Chun-Chen Tu. Generating Contrastive Explanations with Monotonic Attribute Functions. CoRR abs/1905.13565. 2019.
           <https://arxiv.org/pdf/1905.12698.pdf>`_

    Args:
        model: Trained classifier, such as a ScikitClassifier that implements
            a predict() and a predict_proba() methods.
        x (numpy.ndarray): row of data.
        coefs (numpy.ndarray): coefficients (weights) corresponding to attribute importance.
        base ((numpy.ndarray): base (default) values of attributes

    Returns:
        bool: True if the relationship is monotonic.
    """
    #find predicted class
    pred_class = model.predict(x.reshape(1,-1))[0]

    x_copy = base.copy()

    #find indexs of coefficients in increasing order of value
    ar = np.argsort(coefs)
    pred_probs = np.zeros(x.shape[0])
    for ind in np.nditer(ar):
        x_copy[ind] = x[ind]
        x_copy_pr = model.predict_proba(x_copy.reshape(1,-1))
        pred_probs[ind] = x_copy_pr[0][pred_class]

    return np.all(np.diff(pred_probs[ar]) >= 0)



import json

import numpy as np
from numpy.testing import assert_allclose

import pandas as pd
from pandas.util.testing import assert_frame_equal

import unittest

from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score

from aix360.algorithms.rbm import FeatureBinarizer, GLRMExplainer, LogisticRuleRegression


class TestLogisticRuleRegression(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.bc = load_breast_cancer()

    def test_classification(self):
        bc_df = pd.DataFrame(self.bc.data, columns=self.bc.feature_names)
        X_train, X_test, Y_train, Y_test = train_test_split(bc_df, self.bc.target, test_size = 0.2, random_state = 31)
        fb = FeatureBinarizer(negations=True)
        X_train_fb = fb.fit_transform(X_train)
        X_test_fb = fb.transform(X_test)

        self.assertEqual(len(X_train_fb.columns), 540)
        self.assertEqual(len(X_test_fb.columns), 540)

        logistic_model = LogisticRuleRegression(maxSolverIter=2000)
        explainer = GLRMExplainer(logistic_model)
        explainer.fit(X_train_fb, Y_train)
        Y_pred = explainer.predict(X_test_fb)

        self.assertGreater(accuracy_score(Y_test, Y_pred), 0.85)
        self.assertGreater(precision_score(Y_test, Y_pred), 0.85)
        self.assertGreater(recall_score(Y_test, Y_pred), 0.85)
        self.assertGreater(f1_score(Y_test, Y_pred), 0.9)

        explanation = explainer.explain()
        expected = pd.DataFrame(columns=["rule", "coefficient"], data=[
          ['(intercept)', -11.2],
          ['worst perimeter <= 116.46 AND worst concave points <= 0.15', -11.9],
          ['worst concave points <= 0.15', 10.1],
          ['worst perimeter <= 116.46 AND worst concave points <= 0.18', 9.8],
          ['worst area <= 930.88', 5.4],
          ['worst area > 680.60 AND worst concavity > 0.22', -3.3],
          ['worst perimeter <= 116.46 AND worst smoothness <= 0.16', 3.1],
          ['mean concave points <= 0.05', 1.5],
          ['worst concavity <= 0.27', 0.9],
          ['worst concave points <= 0.12', 0.63],
          ['worst perimeter <= 104.38', -0.02]
        ])
        assert_frame_equal(explanation, expected, check_dtype=False, check_exact=False, check_less_precise=1)

        figs, _ = explainer.visualize(bc_df, fb)
        with open('tests/rbm/logistic_plot_data.json') as fp:
            plot_data = json.load(fp)
            for k,v in plot_data.items():
                obtained_plot = figs[k].axes[0].lines[0].get_xydata()
                assert_allclose(np.array(v), obtained_plot, rtol=1e-2)


if __name__ == '__main__':
    unittest.main()



import pandas as pd    
def openDataFile(fileName) :  
    """ Open dataset file and populate X, Y, and E
    Args: 
        fileName (String) : filename of dataset, a structured (CSV) dataset where
                - The first N-2 columns are the features (X).  
                - The next to last column is the label (Y) {0, 1}
                - The last column gives the explanations (E) {0, 1, ..., MaxE}.  We assume the explanation space
                    is dense, i.e., if there are MaxE+1 unique explanations, they will be given IDs from 0 .. MaxE
                - first row contains header information for each column and should be "Y" for labels and "E" for explanations
                - each row is an instance 
    Returns:
            X : list of features vectors
            Y : list of labels
            E : list of explanations
    """                           
    data = pd.read_csv(fileName)     # Load datafile into a dataframe
    X = data.iloc[:,:-2]   # Choose all rows and all cols, except for the last 2 cols
    Y = data['Y']          # Choose col with header 'Y'
    E = data['E']          # Choose col with header 'E'

    return X, Y, E


import numpy as np



def prof_weight_compute(list_probe_filenames,start_layer,last_layer,y_train2):
	"""
	Key function that computes the per sample weights based on probe classifier confidences from various layers. It averages the probe 
	confidences of the layers.

	Parmeters:
	list_probe_filenames (List of strings): A list of strings specifying path where probe confidences for different layers are stored.
	start_layer (int): Index of the starting layer to average with from the list of filenames each storing probe confidences for a layer
	final_layer (int): Index of the last layer to average with from the list of filenames each storing probe confidences for a layer.
	y_train2 (numpy array): Labels corresponding to the set of datasamples in each of the filenames. Dimension - (num of samples x num of classes)
	
	Return:
	prof_weights (numpy array): A vector of weights for each data sample to train the simple model with. Dimension - (num of samples x 1)
	"""
	probe_2_prediction=np.load(list_probe_filenames[start_layer])
	for r in range(start_layer+1,last_layer+1):
		probe_2_prediction=probe_2_prediction+np.load(list_probe_filenames[r])
	##*********************************************
	num_layers_use=last_layer-start_layer+1
	probe_2_prediction=probe_2_prediction.astype(float)/num_layers_use
	num_examples=probe_2_prediction.shape[0]
	prof_weights=np.zeros((num_examples,1))
	##*********
	for r in range(num_examples):
		prof_weights[r]=np.dot(probe_2_prediction[r,:],y_train2[r,:])
	return prof_weights






import json

import numpy as np
from numpy.testing import assert_allclose

import pandas as pd
from pandas.util.testing import assert_frame_equal

import unittest

from sklearn.datasets import load_boston
from sklearn.model_selection import train_test_split
from sklearn.metrics import r2_score, explained_variance_score, mean_absolute_error, max_error

from aix360.algorithms.rbm import FeatureBinarizer, GLRMExplainer, LinearRuleRegression


class TestLinearRuleRegression(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.boston = load_boston()

    def test_classification(self):
        boston_df = pd.DataFrame(self.boston.data, columns=self.boston.feature_names)
        X_train, X_test, Y_train, Y_test = train_test_split(boston_df, self.boston.target, test_size = 0.25, random_state = 31)
        fb = FeatureBinarizer(negations=True)
        X_train_fb = fb.fit_transform(X_train)
        X_test_fb = fb.transform(X_test)

        self.assertEqual(len(X_train_fb.columns), 196)
        self.assertEqual(len(X_test_fb.columns), 196)

        linear_model = LinearRuleRegression()
        explainer = GLRMExplainer(linear_model)
        explainer.fit(X_train_fb, Y_train)
        Y_pred = explainer.predict(X_test_fb)

        self.assertGreater(r2_score(Y_test, Y_pred), 0.8)
        self.assertGreater(explained_variance_score(Y_test, Y_pred), 0.8)
        self.assertLess(mean_absolute_error(Y_test, Y_pred), 3)
        self.assertLess(max_error(Y_test, Y_pred), 12)

        explanation = explainer.explain()
        explanation = explainer.explain()
        expected = pd.DataFrame(columns=["rule", "coefficient"], data=[
            ['(intercept)', 21.9],
            ['NOX <= 0.66', 6.3],
            ['RM <= 7.16 AND DIS > 1.62', -5.8],
            ['LSTAT <= 4.66', 5.5],
            ['DIS <= 3.32 AND RAD > 2.00 AND B > 295.98 AND LSTAT <= 22.79', 4.8],
            ['CHAS not  AND PTRATIO > 16.10', -3.9],
            ['RM <= 7.16 AND RAD <= 6.00', -3.3],
            ['TAX > 293.00 AND LSTAT > 4.66', -2.9],
            ['LSTAT <= 15.03', 2.8],
            ['INDUS > 4.05 AND LSTAT > 4.66', -2.5],
            ['DIS <= 7.24 AND RAD > 2.00 AND PTRATIO <= 20.90 AND B <= 394.99 AND B > 295.98 AND LSTAT <= 22.79', 2.5],
            ['LSTAT <= 9.48', 2.5],
            ['CRIM <= 9.84 AND DIS <= 4.64 AND RAD > 1.00 AND TAX <= 666.00 AND LSTAT <= 22.79', 2.2],
            ['LSTAT <= 17.60', 1.9],
            ['TAX > 330.00 AND LSTAT > 4.66', -1.8],
            ['CRIM <= 9.84 AND CRIM > 0.06 AND PTRATIO <= 20.90', 1.8],
            ['LSTAT <= 6.25', 1.6],
            ['RM <= 7.16 AND B > 380.27', -1.6],
            ['LSTAT <= 11.12', 1.6],
            ['RAD > 2.00 AND LSTAT <= 22.79', 1.2],
            ['RM <= 7.16', -1.2],
            ['CHAS not  AND RM <= 7.16', 1.2],
            ['RM <= 6.51', -1.1],
            ['CRIM <= 9.84 AND DIS <= 3.95 AND TAX <= 666.00 AND PTRATIO <= 20.90 AND B > 295.98', 1.0],
            ['CRIM <= 9.84 AND RAD > 1.00 AND LSTAT <= 22.79', 1.0],
            ['DIS <= 3.95 AND LSTAT <= 22.79', -0.9],
            ['RM <= 6.74', -0.8],
            ['PTRATIO <= 19.52', 0.8],
            ['NOX <= 0.66 AND PTRATIO <= 20.90 AND LSTAT <= 22.79', -0.8],
            ['RAD > 4.00 AND LSTAT <= 22.79', -0.63],
            ['B <= 391.27 AND LSTAT <= 22.79', 0.5],
            ['LSTAT <= 7.58', 0.44],
            ['LSTAT <= 13.14', 0.17]
        ])
        assert_frame_equal(explanation, expected, check_dtype=False, check_exact=False, check_less_precise=1)

        figs, _ = explainer.visualize(boston_df, fb)
        with open('tests/rbm/linear_plot_data.json') as fp:
            plot_data = json.load(fp)
            for k,v in plot_data.items():
                obtained_plot = figs[k].axes[0].lines[0].get_xydata()
                assert_allclose(np.array(v), obtained_plot, rtol=1e-2)


if __name__ == '__main__':
    unittest.main()



from ..dise import DISExplainer


class BRCGExplainer(DISExplainer):
    """
    Boolean Rule Column Generation explainer. Provides access to 
    :class:`aix360.algorithms.rbm.boolean_rule_cg.BooleanRuleCG`, which
    implements a directly interpretable supervised learning method
    for binary classification that learns a Boolean rule in disjunctive
    normal form (DNF) or conjunctive normal form (CNF) using column generation (CG).
    AIX360 implements a heuristic beam search version of BRCG that is less 
    computationally intensive than the published integer programming version [#NeurIPS2018]_.

    References:
        .. [#NeurIPS2018] `S. Dash, O. Günlük, D. Wei, "Boolean decision rules via
           column generation." Neural Information Processing Systems (NeurIPS), 2018.
           <https://papers.nips.cc/paper/7716-boolean-decision-rules-via-column-generation.pdf>`_
    """

    def __init__(self, model):
        """
        Initialize a BRCGExplainer object.

        Args:
          model: model to operate on, instance of :class:`aix360.algorithms.rbm.boolean_rule_cg.BooleanRuleCG`
        """
        super(BRCGExplainer, self).__init__()
        self._model = model

    def set_params(self, *argv, **kwargs):
        """
        Set parameters for the explainer.
        """
        pass

    def fit(self, X_train, Y_train, *argv, **kwargs):
        """Fit model to training data.

        Args:
            X_train (DataFrame): Binarized features with MultiIndex column labels
            Y_train (array): Binary-valued target variable
        Returns:
            BRCGExplainer: Self
        """
        self._model.fit(X_train, Y_train, **kwargs)

    def predict(self, X, *argv, **kwargs):
        """Predict class labels.

        Args:
            X (DataFrame): Binarized features with MultiIndex column labels
        Returns:
            array: y -- Predicted labels
        """
        return self._model.predict(X, **kwargs)

    def explain(self, *argv, **kwargs):
        """Return rules comprising the underlying model.

        Args:
            maxConj (int, optional): Maximum number of conjunctions to show
            prec (int, optional): Number of decimal places to show for floating-value thresholds
        Returns:
            Dictionary containing

            * isCNF (bool): flag signaling whether model is CNF or DNF
            * rules (list): selected conjunctions formatted as strings
        """
        return self._model.explain(**kwargs)



import numpy as np
import pandas as pd


class PricingInstance(object):
    """Instance of the pricing problem"""
    def __init__(self, rp, rn, Xp, Xn, v0, z0):
        self.rp = rp
        self.rn = rn
        self.Xp = Xp
        self.Xn = Xn
        self.v0 = v0
        self.z0 = z0
    
    def eval_singletons(self, lambda1):
        """Evaluate all singleton solutions (adding one literal)"""
        self.Rp = np.dot(self.rp, self.Xp)
        self.Rn = np.dot(self.rn, self.Xn)
        self.v1 = self.v0 - self.Rp - self.Rn + lambda1
        self.v1 = pd.Series(self.v1, index=self.Xp.columns)
        
    def compute_LB(self, lambda1):
        """Compute lower bound on higher-order solutions"""
        Rp0 = self.rp.sum()
        if np.ndim(lambda1):
            self.LB = np.array([])
        else:
            self.LB = np.minimum(np.cumsum(np.sort(self.Rp)[::-1])[1:], Rp0)
            self.LB += np.sort(self.Rn)[-2::-1]
            self.LB -= lambda1 * np.arange(2, len(self.Rp)+1)
            self.LB = self.v0 - self.LB

        # Lower bound specific to each singleton solution
        self.LB1 = self.v1 + self.Rp - Rp0 + lambda1
        if len(self.LB):
            self.LB1[self.LB1 < self.LB.min()] = self.LB.min()


def beam_search_K1(r, X, lambda0, lambda1, UB=0, D=10, B=5, wLB=0.5, eps=1e-6, stopEarly=True):
    """Beam search to generate SINGLE SOLUTION (K = 1) to pricing problem
    Problem parameters:
        r = cost vector (residuals)
        X = binary feature DataFrame
        lambda0 = fixed cost of a term
        lambda1 = cost per literal
    
    Algorithm parameters:
        UB = initial upper bound on value of solutions
        D = maximum degree
        B = beam width
        wLB = weight on lower bound in evaluating nodes
        eps = numerical tolerance on comparisons
        stopEarly = stop after current degree once solution is found
    """

    # Initialize output
    vOut = np.array([])
    zOut = pd.Series(index=X.columns)

    # Initialize queue with root instance
    # Separate data according to positive and negative residuals
    rp = r[r > 0]
    rn = r[r < 0]
    Xp = 1 - X.loc[r > 0]
    Xn = 1 - X.loc[r < 0]
    instCurr = [PricingInstance(rp, rn, Xp, Xn, r.sum() + lambda0, pd.Series(0, index=zOut.index))]
    
    # Iterate over increasing degree while queue is non-empty
    deg = 0
    while (not len(vOut) or not stopEarly) and len(instCurr) and deg < D:
        deg += 1

        # Initialize list of children to process
        vNext = np.array([])
        vNextMax = np.inf
        zNext = pd.DataFrame([], index=X.columns)
        idxInstNext = np.array([], dtype=int)
        idxFeatNext = np.array([])
        
        # Process instances in queue
        for (idxInst, inst) in enumerate(instCurr):

            # Evaluate all singleton solutions
            inst.eval_singletons(lambda1)

            # Solutions that improve on current output
            vCand = inst.v1[inst.v1 < UB - eps]
            if len(vCand):
                # Update output with best of these solutions
                idxMin = vCand.idxmin()
                UB = vCand[idxMin]
                vOut = np.array([UB])
                zOut = inst.z0.copy()
                zOut[idxMin] = 1
                
            # Compute lower bounds on higher-degree solutions
            inst.compute_LB(lambda1)    
            # Evaluate children using weighted average of their costs and LBs
            vChild = (1 - wLB) * inst.v1 + wLB * inst.LB1

            # Best children with potential to improve on current output and current candidates (allow for duplicate removal)
            vChild = vChild[(inst.LB1 < UB - eps) & (vChild < vNextMax - eps)].sort_values()[:B+idxInst]
            if len(vChild):
                # Feature indicators of these best children
                zChild = pd.DataFrame(zOut.index.values[:,np.newaxis] == vChild.index.values, index=zOut.index).astype(int)
                zChild = zChild.add(inst.z0, axis=0)
                
                # Append to current candidates
                vNext = np.append(vNext, vChild.values)
                zNext = pd.concat([zNext, zChild], axis=1, ignore_index=True)
                idxInstNext = np.append(idxInstNext, np.full(B+idxInst, idxInst))
                idxFeatNext = np.append(idxFeatNext, vChild.index.values)
                # Remove duplicates
                _, idxUniq = np.unique(zNext, return_index=True, axis=1)
                vNext = vNext[idxUniq]
                zNext = zNext.iloc[:,idxUniq]
                idxInstNext = idxInstNext[idxUniq]
                idxFeatNext = idxFeatNext[idxUniq]
                # Update candidates
                idxBest = np.argsort(vNext)[:B]
                vNext = vNext[idxBest]
                if len(vNext):
                    vNextMax = vNext[-1]
                zNext = zNext.iloc[:,idxBest]
                zNext.columns = range(zNext.shape[1])
                idxInstNext = idxInstNext[idxBest]
                idxFeatNext = idxFeatNext[idxBest]
                
        # Instances to process in next iteration
        instNext = []
        for (idxInst, i, idxz) in zip(idxInstNext, idxFeatNext, zNext):
            # Create pricing instance
            # Remove covered rows
            rowKeep = instCurr[idxInst].Xp[i] == 0
            rp = instCurr[idxInst].rp[rowKeep]
            Xp = instCurr[idxInst].Xp.loc[rowKeep]
            rowKeep = instCurr[idxInst].Xn[i] == 0
            rn = instCurr[idxInst].rn[rowKeep]
            Xn = instCurr[idxInst].Xn.loc[rowKeep]
            # Remove redundant features
            if type(Xp.columns) is pd.MultiIndex: 
                colKeep = pd.Series(Xp.columns.get_level_values(0) != i[0], index=Xp.columns)
                if i[1] == '<=':
                    thresh = Xp[i[0]].columns.get_level_values(1).to_series().replace('NaN', np.nan)
                    colKeep[i[0]] = (Xp[i[0]].columns.get_level_values(0) == '>') & (thresh < i[2])
                elif i[1] == '>':
                    thresh = Xp[i[0]].columns.get_level_values(1).to_series().replace('NaN', np.nan)
                    colKeep[i[0]] = (Xp[i[0]].columns.get_level_values(0) == '<=') & (thresh > i[2])
                elif i[1] == '!=':
                    colKeep[i[0]] = (Xp[i[0]].columns.get_level_values(0) == '!=') & (Xp[i[0]].columns.get_level_values(1) != i[2])
                Xp = Xp.loc[:, colKeep]
                Xn = Xn.loc[:, colKeep]
            instNext.append(PricingInstance(rp, rn, Xp, Xn, instCurr[idxInst].v1[i], zNext[idxz]))

        instCurr = instNext
        
    # Conjunctions corresponding to solutions
    if zOut.count():
        zOut = pd.DataFrame(zOut)
    else:
        zOut = pd.DataFrame(index=X.columns)

    aOut = 1 - (np.dot(1 - X, zOut) > 0)
    
    return vOut, zOut, aOut


def beam_search(r, X, lambda0, lambda1, K=1, UB=0, D=10, B=5, wLB=0.5, eps=1e-6, stopEarly=False):
    """Beam search to generate solutions to pricing problem
    Problem parameters:
        r = cost vector (residuals)
        X = binary feature DataFrame
        lambda0 = fixed cost of a term
        lambda1 = cost per literal
    
    Algorithm parameters:
        K = maximum number of solutions returned
        UB = initial upper bound on value of solutions
        D = maximum degree
        B = beam width
        wLB = weight on lower bound in evaluating nodes
        eps = numerical tolerance on comparisons
        stopEarly = stop after current degree once solution is found
    """

    # Initialize output
    vOut = np.array([])
    zOut = pd.DataFrame(index=X.columns)

    # Remove redundant rows by grouping by unique feature combinations and summing residual
    X2 = X.copy()
    r2 = r
    
    # Initialize queue with root instance
    # Separate data according to positive and negative residuals
    rp = r2[r2 > 0]
    rn = r2[r2 < 0]
    Xp = 1 - X2.loc[r2 > 0]
    Xn = 1 - X2.loc[r2 < 0]
    instCurr = [PricingInstance(rp, rn, Xp, Xn, r2.sum() + lambda0, pd.Series(0, index=zOut.index))]
    
    # Iterate over increasing degree while queue is non-empty
    deg = 0
    while (not len(vOut) or not stopEarly) and len(instCurr) and deg < D:
        deg += 1

        # Initialize list of children to process
        vNext = np.array([])
        vNextMax = np.inf
        zNext = pd.DataFrame([], index=X2.columns)
        idxInstNext = np.array([], dtype=int)
        idxFeatNext = np.array([])
        
        # Process instances in queue
        for (idxInst, inst) in enumerate(instCurr):

            # Evaluate all singleton solutions
            inst.eval_singletons(lambda1)

            # Best solutions that also improve on current output (allow for duplicate removal)
            vCand = inst.v1[inst.v1 < UB - eps].sort_values()[:K+B]
            if len(vCand):
                zCand = pd.DataFrame(zOut.index.values[:,np.newaxis] == vCand.index.values, index=zOut.index).astype(int)
                zCand = zCand.add(inst.z0, axis=0)
                # Append to current output
                vOut = np.append(vOut, vCand.values)
                zOut = pd.concat([zOut, zCand], axis=1, ignore_index=True)
                # Remove duplicates
                _, idxUniq = np.unique(zOut, return_index=True, axis=1)
                vOut = vOut[idxUniq]
                zOut = zOut.iloc[:,idxUniq]
                # Update output
                indBest = np.argsort(vOut)[:K]
                vOut = vOut[indBest]
                UB = vOut.max()
                zOut = zOut.iloc[:,indBest]
                zOut.columns = range(zOut.shape[1])
            
            # Compute lower bounds on higher-degree solutions
            inst.compute_LB(lambda1)
            # Evaluate children using weighted average of their costs and LBs
            vChild = (1 - wLB) * inst.v1 + wLB * inst.LB1
            
            # Best children with potential to improve on current output and current candidates (allow for duplicate removal)
            vChild = vChild[(inst.LB1 < UB - eps) & (vChild < vNextMax - eps)].sort_values()[:B+idxInst]
            if len(vChild):
                # Feature indicators of these best children
                zChild = pd.DataFrame(zOut.index.values[:,np.newaxis] == vChild.index.values, index=zOut.index).astype(int)
                zChild = zChild.add(inst.z0, axis=0)
                
                # Append to current candidates
                vNext = np.append(vNext, vChild.values)
                zNext = pd.concat([zNext, zChild], axis=1, ignore_index=True)
                idxInstNext = np.append(idxInstNext, np.full(B+idxInst, idxInst))
                idxFeatNext = np.append(idxFeatNext, vChild.index.values)
                # Remove duplicates
                _, idxUniq = np.unique(zNext, return_index=True, axis=1)
                vNext = vNext[idxUniq]
                zNext = zNext.iloc[:,idxUniq]
                idxInstNext = idxInstNext[idxUniq]
                idxFeatNext = idxFeatNext[idxUniq]
                # Update candidates
                idxBest = np.argsort(vNext)[:B]
                vNext = vNext[idxBest]
                if len(vNext):
                    vNextMax = vNext[-1]
                zNext = zNext.iloc[:,idxBest]
                zNext.columns = range(zNext.shape[1])
                idxInstNext = idxInstNext[idxBest]
                idxFeatNext = idxFeatNext[idxBest]
                
        # Instances to process in next iteration
        instNext = []
        for (idxInst, i, idxz) in zip(idxInstNext, idxFeatNext, zNext):
            # Create pricing instance
            # Remove covered rows
            rowKeep = instCurr[idxInst].Xp[i] == 0
            rp = instCurr[idxInst].rp[rowKeep]
            Xp = instCurr[idxInst].Xp.loc[rowKeep]
            rowKeep = instCurr[idxInst].Xn[i] == 0
            rn = instCurr[idxInst].rn[rowKeep]
            Xn = instCurr[idxInst].Xn.loc[rowKeep]
            # Remove redundant features
            colKeep = pd.Series(Xp.columns.get_level_values(0) != i[0], index=Xp.columns)
            if i[1] == '<=':
                thresh = Xp[i[0]].columns.get_level_values(1).to_series().replace('NaN', np.nan)
                colKeep[i[0]] = (Xp[i[0]].columns.get_level_values(0) == '>') & (thresh < i[2])
            elif i[1] == '>':
                thresh = Xp[i[0]].columns.get_level_values(1).to_series().replace('NaN', np.nan)
                colKeep[i[0]] = (Xp[i[0]].columns.get_level_values(0) == '<=') & (thresh > i[2])
            elif i[1] == '!=':
                colKeep[i[0]] = (Xp[i[0]].columns.get_level_values(0) == '!=') & (Xp[i[0]].columns.get_level_values(1) != i[2])
            Xp = Xp.loc[:, colKeep]
            Xn = Xn.loc[:, colKeep]
            instNext.append(PricingInstance(rp, rn, Xp, Xn, instCurr[idxInst].v1[i], zNext[idxz]))

        instCurr = instNext
        
    # Conjunctions corresponding to solutions
    aOut = 1 - (np.dot(1 - X, zOut) > 0)

    return vOut, zOut, aOut



## aen_attack.py -- attack a network optimizing elastic-net distance with an en decision rule
##                  when autoencoder loss is applied
##
## Copyright (C) 2018, PaiShun Ting <paishun@umich.edu>
##                     Chun-Chen Tu <timtu@umich.edu>
##                     Pin-Yu Chen <Pin-Yu.Chen@ibm.com>
## Copyright (C) 2017, Yash Sharma <ysharma1126@gmail.com>.
## Copyright (C) 2016, Nicholas Carlini <nicholas@carlini.com>.
##
## This program is licenced under the BSD 2-Clause licence,
## contained in the LICENCE file in this directory.

import sys, os
import tensorflow as tf
import numpy as np
from tensorflow.contrib.keras.api.keras.models import Model, Sequential, model_from_json
from tensorflow.contrib.keras.api.keras.callbacks import ModelCheckpoint

class AEADEN:
    def __init__(self, sess, model, mask_mat, mode, batch_size, kappa, init_learning_rate,
                 binary_search_steps, max_iterations, initial_const, beta, gamma, attributes, aix360_path):
        """
        Initialize PP explainer object. 
        
        Args:
            sess (tensorflow.python.client.session.Session): Tensorflow session
            model: KerasClassifier that contains a trained model to be explained
            mask_mat (numpy.ndarry): Array containing PP masks for each class
            mode (str): "PN" for pertinent negative or "PP" for pertinent positive
            batch_size (int): batch size for how many instances to explain
            kappa (float): Confidence parameter that controls difference between prediction of
                PN (or PP) and original prediction
            init_learning_rate (float): initial learning rate for gradient descent optimizer
            binary_search_steps (int): Controls number of random restarts to find best PN
            max_iterations (int): Max number iterations to run some version of gradient descent on
                PP optimization problem from a single random initialization, i.e., total 
                number of iterations wll be arg_binary_search_steps * arg_max_iterations
            initial_const (int): Constant used for upper/lower bounds in binary search
            gamma (float): Penalty parameter encouraging addition of attributes for PP
            attributes (str list): list of attributes to load attribute classifiers for
            aix360_path (str): path to aix360 used to determine paths to pretrained attribute classifiers 
        """

#        image_size, num_channels, nun_classes = model.image_size, model.num_channels, model.num_labels
        # %%change%%
        image_size = model._input_shape[0]
        num_channels = model._input_shape[2]
        nun_classes = model._nb_classes 
        shape = (batch_size, image_size, image_size, num_channels)
        mask_shape = (batch_size, image_size, image_size, 1)
        mask_num = mask_mat.shape[0]
        mask_vec_shape = (mask_num, 1, 1)
        mask_mat_shape = (mask_num, image_size, image_size)
        self.mask_num = mask_num
        self.sess = sess
        self.INIT_LEARNING_RATE = init_learning_rate
        self.MAX_ITERATIONS = max_iterations
        self.BINARY_SEARCH_STEPS = binary_search_steps
        self.kappa = kappa
        self.init_const = initial_const
        self.batch_size = batch_size
        self.AE = None
        self.mode = mode
        self.beta = beta
        self.gamma = gamma
        self.attributes = attributes
        self.aix360_path = aix360_path
        
        ### Load attribute classifier
        nn_type = "simple"
        #import copy
        attr_model_list=[]
        for attr in self.attributes:
            # load test data into memory using Image Data Generator
#            print("Loading data for " + attr + " into memory")
            # load json and create model
            json_file_name = os.path.join(aix360_path, "models/CEM_MAF/{}_{}_model.json".format(nn_type, attr))
            json_file = open(json_file_name, 'r')
            loaded_model_json = json_file.read()
            json_file.close()
            loaded_model = model_from_json(loaded_model_json)
            # load weights into new model
            weight_file_name = os.path.join(aix360_path, "models/CEM_MAF/{}_{}_weights.h5".format(nn_type, attr))
            loaded_model.load_weights(weight_file_name)
            print("Loaded model for " + attr + " from disk")
            attr_model_list.append(loaded_model)
        
        print("# of attr models is",len(attr_model_list))


#        print("beta:{}".format(self.beta))
        # these are variables to be more efficient in sending data to tf
        self.orig_img = tf.Variable(np.zeros(shape), dtype=tf.float32)
        self.mask_vec = tf.Variable(np.zeros(mask_vec_shape), dtype=tf.float32)
        self.mask_vec_s = tf.Variable(np.zeros(mask_vec_shape), dtype=tf.float32)
        self.mask_mat = tf.constant(mask_mat, dtype=tf.float32)
        # self.img_mask = tf.Variable(np.zeros(mask_shape), dtype=tf.float32)
        # self.img_mask_s = tf.Variable(np.zeros(mask_shape), dtype=tf.float32, name="var_mask_s")
        self.target_lab = tf.Variable(np.zeros((batch_size,nun_classes)), dtype=tf.float32)
        self.const = tf.Variable(np.zeros(batch_size), dtype=tf.float32)
        self.global_step = tf.Variable(0.0, trainable=False)

        # and here's what we use to assign them
        self.assign_orig_img = tf.placeholder(tf.float32, shape)
        self.assign_mask_vec = tf.placeholder(tf.float32, mask_vec_shape)
        self.assign_mask_vec_s = tf.placeholder(tf.float32, mask_vec_shape)
        # self.assign_img_mask = tf.placeholder(tf.float32, mask_shape)
        # self.assign_img_mask_s = tf.placeholder(tf.float32, mask_shape)
        self.assign_target_lab = tf.placeholder(tf.float32, (batch_size,nun_classes))
        self.assign_const = tf.placeholder(tf.float32, [batch_size])


        """Fast Iterative Soft Thresholding"""
        """--------------------------------"""
        

        # self.zt = tf.divide(self.global_step, self.global_step+tf.cast(3, tf.float32))
        # cond1 = tf.cast(tf.greater(tf.subtract(self.adv_img_mask_s, self.orig_img),self.beta), tf.float32)
        # cond2 = tf.cast(tf.less_equal(tf.abs(tf.subtract(self.adv_img_s,self.orig_img)),self.beta), tf.float32)
        # cond3 = tf.cast(tf.less(tf.subtract(self.adv_img_s, self.orig_img),tf.negative(self.beta)), tf.float32)
        # upper = tf.minimum(tf.subtract(self.adv_img_s, self.beta), tf.cast(0.5, tf.float32))
        # lower = tf.maximum(tf.add(self.adv_img_s, self.beta), tf.cast(-0.5, tf.float32))
        # self.assign_adv_img = tf.multiply(cond1,upper)+tf.multiply(cond2,self.orig_img)+tf.multiply(cond3,lower)

        self.zt = tf.divide(self.global_step, self.global_step+tf.cast(3, tf.float32))
        """
        x = x - beta if x > beta
        x = 0 if x < beta
        """
        cond1 = tf.cast(tf.greater(self.mask_vec_s, self.beta), tf.float32)
        cond2 = tf.cast(tf.less_equal(self.mask_vec_s, self.beta), tf.float32)
        upper = tf.minimum(tf.subtract(self.mask_vec_s, self.beta), tf.cast(1, tf.float32))
        self.assign_mask_vec = tf.multiply(cond1,upper) + tf.multiply(cond2, tf.constant(0, tf.float32))
        self.assign_mask_vec_s = self.assign_mask_vec+tf.multiply(self.zt, self.assign_mask_vec-self.mask_vec)
        # cond4=tf.cast(tf.greater(tf.subtract( self.assign_adv_img, self.orig_img),0), tf.float32)
        # cond5=tf.cast(tf.less_equal(tf.subtract( self.assign_adv_img,self.orig_img),0), tf.float32)
        # if self.mode == "PP":
        #     self.assign_adv_img = tf.multiply(cond5,self.assign_adv_img)+tf.multiply(cond4,self.orig_img)
        # elif self.mode == "PN":
        #     self.assign_adv_img = tf.multiply(cond4,self.assign_adv_img)+tf.multiply(cond5,self.orig_img)

        # self.assign_adv_img_s = self.assign_adv_img+tf.multiply(self.zt, self.assign_adv_img-self.adv_img)
        # cond6=tf.cast(tf.greater(tf.subtract(  self.assign_adv_img_s, self.orig_img),0), tf.float32)
        # cond7=tf.cast(tf.less_equal(tf.subtract(  self.assign_adv_img_s,self.orig_img),0), tf.float32)
        # if self.mode == "PP":
        #     self.assign_adv_img_s = tf.multiply(cond7, self.assign_adv_img_s)+tf.multiply(cond6,self.orig_img)
        # elif self.mode == "PN":
        #     self.assign_adv_img_s = tf.multiply(cond6, self.assign_adv_img_s)+tf.multiply(cond7,self.orig_img)
        self.mask_updater = tf.assign(self.mask_vec, self.assign_mask_vec)
        self.mask_updater_s = tf.assign(self.mask_vec_s, self.assign_mask_vec_s)
        """ Thresholding """

        # mask_ones = tf.constant(1, tf.float32)
        # mask_zeros = tf.constant(0, tf.float32)

        # mask_cond1 = tf.cast(tf.greater(self.img_mask, 0.5), tf.float32)
        # mask_cond2 = tf.cast(tf.less_equal(self.img_mask, 0.5), tf.float32)
        # self.img_mask_threshold = tf.multiply(mask_cond1, mask_ones)+tf.multiply(mask_cond2, mask_zeros)
        # cannot find gradient
        self.img_mask = tf.reduce_sum(self.mask_vec * self.mask_mat, axis=0)
        self.img_mask = tf.expand_dims(self.img_mask, axis=2)
        self.img_mask = tf.expand_dims(self.img_mask, axis=0)
        self.adv_img = tf.multiply(self.img_mask, self.orig_img)


        # mask_s_cond1 = tf.cast(tf.greater(self.img_mask_s, 0.5), tf.float32)
        # mask_s_cond2 = tf.cast(tf.less_equal(self.img_mask_s, 0.5), tf.float32)
        # self.img_mask_threshold_s = tf.multiply(mask_s_cond1, mask_ones)+tf.multiply(mask_s_cond2, mask_zeros)

        # cannot find gradient
        self.img_mask_s = tf.reduce_sum(self.mask_vec_s * self.mask_mat, axis=0)
        self.img_mask_s = tf.expand_dims(self.img_mask_s, axis=2)
        self.img_mask_s = tf.expand_dims(self.img_mask_s, axis=0)
        self.adv_img_s = tf.multiply(self.img_mask_s, self.orig_img)

        
        """--------------------------------"""
        # prediction from attribute classifer
        self.delta_img = self.orig_img-self.adv_img
        self.delta_img_s = self.orig_img-self.adv_img_s
        if self.mode == "PP":
            self.attr_score = tf.constant(0, dtype="float32")
            self.attr_score_s = tf.constant(0, dtype="float32")
            #print(attr_model_list[0].predict(self.adv_img)) 
            #print(loaded_model.predict(self.adv_img)) 
            for i in range(len(attr_model_list)):
                self.attr_score = self.attr_score + tf.maximum(attr_model_list[i](self.adv_img) - attr_model_list[i](self.orig_img),tf.constant(0, tf.float32))
                self.attr_score_s = self.attr_score_s + tf.maximum(attr_model_list[i](self.adv_img_s) - attr_model_list[i](self.orig_img),tf.constant(0, tf.float32))
                #print(self.attr_score.shape) 
            self.attr_score = tf.squeeze(self.attr_score)
            self.attr_score_s = tf.squeeze(self.attr_score_s)  
           #self.ImgToEnforceLabel_Score = model.predict(self.adv_img)
            #self.ImgToEnforceLabel_Score_s = model.predict(self.adv_img_s)
# %%change%%
        elif self.mode == "PN":
#            self.ImgToEnforceLabel_Score = model.predict(self.adv_img)
#            self.ImgToEnforceLabel_Score_s = model.predict(self.adv_img_s)
            self.ImgToEnforceLabel_Score = model.predictsym(self.adv_img)
            self.ImgToEnforceLabel_Score_s = model.predictsym(self.adv_img_s)

        # prediction BEFORE-SOFTMAX of the model
        self.delta_img = self.orig_img-self.adv_img
        self.delta_img_s = self.orig_img-self.adv_img_s
        if self.mode == "PP":
#            self.ImgToEnforceLabel_Score = model.predict(self.adv_img)
#            self.ImgToEnforceLabel_Score_s = model.predict(self.adv_img_s)
            self.ImgToEnforceLabel_Score = model.predictsym(self.adv_img)
            self.ImgToEnforceLabel_Score_s = model.predictsym(self.adv_img_s)
        elif self.mode == "PN":
#            self.ImgToEnforceLabel_Score = model.predict(self.adv_img)
#            self.ImgToEnforceLabel_Score_s = model.predict(self.adv_img_s)
            self.ImgToEnforceLabel_Score = model.predictsym(self.adv_img)
            self.ImgToEnforceLabel_Score_s = model.predictsym(self.adv_img_s)

        # distance to the input data
        self.L2_dist = tf.reduce_sum(tf.square(self.img_mask),[1,2,3])
        self.L2_dist_s = tf.reduce_sum(tf.square(self.img_mask_s),[1,2,3])
        self.L1_dist = tf.reduce_sum(tf.abs(self.img_mask),[1,2,3])
        self.L1_dist_s = tf.reduce_sum(tf.abs(self.img_mask_s),[1,2,3])
        self.EN_dist = self.L2_dist + tf.multiply(self.L1_dist, self.beta)
        self.EN_dist_s = self.L2_dist_s + tf.multiply(self.L1_dist_s, self.beta)

        # compute the probability of the label class versus the maximum other
        self.target_lab_score        = tf.reduce_sum((self.target_lab)*self.ImgToEnforceLabel_Score,1)
        target_lab_score_s           = tf.reduce_sum((self.target_lab)*self.ImgToEnforceLabel_Score_s,1)
        self.max_nontarget_lab_score = tf.reduce_max((1-self.target_lab)*self.ImgToEnforceLabel_Score - (self.target_lab*10000),1)
        max_nontarget_lab_score_s    = tf.reduce_max((1-self.target_lab)*self.ImgToEnforceLabel_Score_s - (self.target_lab*10000),1)
        if self.mode == "PP":
            Loss_Attack = tf.maximum(0.0, self.max_nontarget_lab_score - self.target_lab_score + self.kappa)
            Loss_Attack_s = tf.maximum(0.0, max_nontarget_lab_score_s - target_lab_score_s + self.kappa)
        elif self.mode == "PN":
            Loss_Attack = tf.maximum(0.0, -self.max_nontarget_lab_score + self.target_lab_score + self.kappa)
            Loss_Attack_s = tf.maximum(0.0, -max_nontarget_lab_score_s + target_lab_score_s + self.kappa)
        # sum up the losses
        self.Loss_L1Dist    = tf.reduce_sum(self.L1_dist)
        self.Loss_L1Dist_s  = tf.reduce_sum(self.L1_dist_s)
        self.Loss_L2Dist    = tf.reduce_sum(self.L2_dist)
        self.Loss_L2Dist_s  = tf.reduce_sum(self.L2_dist_s)
        self.Loss_Attack    = tf.reduce_sum(self.const*Loss_Attack)
        with tf.name_scope("loss_attack_s"):
            self.Loss_Attack_s  = tf.reduce_sum(self.const*Loss_Attack_s)
        if self.AE:
            if self.mode == "PP":
                self.Loss_AE_Dist   = self.gamma*tf.square(tf.norm(self.AE(self.delta_img)-self.delta_img))
                self.Loss_AE_Dist_s = self.gamma*tf.square(tf.norm(self.AE(self.delta_img)-self.delta_img_s))
            elif self.mode == "PN":
                self.Loss_AE_Dist   = self.gamma*tf.square(tf.norm(self.AE(self.adv_img)-self.adv_img))
                self.Loss_AE_Dist_s = self.gamma*tf.square(tf.norm(self.AE(self.adv_img_s)-self.adv_img_s))
        else:
            self.Loss_AE_Dist = tf.constant(0, dtype="float32")
            self.Loss_AE_Dist_s = tf.constant(0, dtype="float32")

        with tf.name_scope("loss"):
            # self.Loss_ToOptimize = self.Loss_Attack_s + self.Loss_L2Dist_s + self.Loss_AE_Dist_s
            self.Loss_ToOptimize = self.Loss_Attack_s + tf.multiply(self.gamma,self.attr_score_s)
        #self.Loss_Overall    = self.Loss_Attack   + self.Loss_L2Dist   + self.Loss_AE_Dist   + tf.multiply(self.beta, self.Loss_L1Dist)
        self.Loss_Overall    = self.Loss_Attack  + tf.multiply(self.gamma,self.attr_score_s)   + tf.multiply(self.beta, self.Loss_L1Dist)
        #print(self.Loss_Attack.shape, self.Loss_Overall.shape)
        self.learning_rate = tf.train.polynomial_decay(self.INIT_LEARNING_RATE, self.global_step, self.MAX_ITERATIONS, 0, power=0.5)
        optimizer = tf.train.GradientDescentOptimizer(self.learning_rate)
        start_vars = set(x.name for x in tf.global_variables())
        # self.train = optimizer.minimize(self.Loss_ToOptimize, var_list=[self.adv_img_s], global_step=self.global_step)
        self.train = optimizer.minimize(self.Loss_ToOptimize, var_list=[self.mask_vec_s], global_step=self.global_step)
        end_vars = tf.global_variables()
        new_vars = [x for x in end_vars if x.name not in start_vars]

        # these are the variables to initialize when we run
        self.setup = []
        self.setup.append(self.orig_img.assign(self.assign_orig_img))
        self.setup.append(self.target_lab.assign(self.assign_target_lab))
        self.setup.append(self.const.assign(self.assign_const))
        self.setup.append(self.mask_vec.assign(self.assign_mask_vec))
        self.setup.append(self.mask_vec_s.assign(self.assign_mask_vec_s))

        self.init = tf.variables_initializer(var_list=[self.global_step]+[self.mask_vec_s]+[self.mask_vec]+new_vars)

    def attack(self, imgs, labs):
        """
        Find PN for an input instance input_image e.g. celebA is shape (1, 224, 224, 3)
        
        Input:
            imgs (numpy.ndarry): images to be explained, of shape (num_images, size, size, channels)
            labs: one hot encoded vectors of target label for original image prediction
            latent (numpy.ndarry): image to be explained, of shape (1, size, size, channels)
                in the latent space
                
        Output: 
            adv_img (numpy.ndarry): the pertinent positive image
        """
        
        def compare(x,y):
            if not isinstance(x, (float, int, np.int64)):
                x = np.copy(x)
                # x[y] -= self.kappa if self.PP else -self.kappa
                if self.mode == "PP":
                    x[y] -= self.kappa
                elif self.mode == "PN":
                    x[y] += self.kappa
                x = np.argmax(x)
            if self.mode == "PP":
                return x==y
            else: 
                return x!=y

        batch_size = self.batch_size

        # set the lower and upper bounds accordingly
        Const_LB = np.zeros(batch_size)
        CONST = np.ones(batch_size)*self.init_const
        Const_UB = np.ones(batch_size)*1e10
        # the best l2, score, and image attack
        overall_best_dist = [1e10]*batch_size
        overall_best_attack = [np.zeros(imgs[0].shape)]*batch_size
        overall_best_mask_vec = [np.zeros((self.mask_num, 1, 1))]*batch_size

        for binary_search_steps_idx in range(self.BINARY_SEARCH_STEPS):
            # completely reset adam's internal state.
            self.sess.run(self.init)
            img_batch = imgs[:batch_size]
            label_batch = labs[:batch_size]
            img_shape = img_batch.shape
            img_mask_vec = np.ones((self.mask_num, 1, 1))

            current_step_best_dist = [1e10]*batch_size
            current_step_best_score = [-1]*batch_size

            # set the variables so that we don't have to send them over again
            self.sess.run(self.setup, {self.assign_orig_img: img_batch,
                                       self.assign_target_lab: label_batch,
                                       self.assign_const: CONST,
                                       self.assign_mask_vec: img_mask_vec,
                                       self.assign_mask_vec_s: img_mask_vec})



            for iteration in range(self.MAX_ITERATIONS):
                # perform the attack
                self.sess.run([self.train])
                self.sess.run([self.mask_updater, self.mask_updater_s])

                Loss_Overall, Loss_EN, OutputScore, adv_img, img_mask = self.sess.run([self.Loss_Overall, self.EN_dist, self.ImgToEnforceLabel_Score, self.adv_img, self.mask_vec])
                # print("max:{}, min:{}".format(np.max(img_mask), np.min(img_mask)))
                Loss_Attack, Loss_L2Dist, Loss_L1Dist, Loss_AE_Dist, Loss_attr = self.sess.run([self.Loss_Attack, self.Loss_L2Dist, self.Loss_L1Dist, self.Loss_AE_Dist, self.attr_score])
                target_lab_score, max_nontarget_lab_score_s = self.sess.run([self.target_lab_score, self.max_nontarget_lab_score])
                if iteration%(self.MAX_ITERATIONS//10) == 0:
                    print("iter:{} const:{}". format(iteration, CONST))
                    print("Loss_Overall:{:.4f}, Loss_Attack:{:.4f}, Loss_attr:{:.4f}". format(Loss_Overall, Loss_Attack, Loss_attr))
                    print("Loss_L2Dist:{:.4f}, Loss_L1Dist:{:.4f}, AE_loss:{}". format(Loss_L2Dist, Loss_L1Dist, Loss_AE_Dist))
                    print("target_lab_score:{:.4f}, max_nontarget_lab_score:{:.4f}". format(target_lab_score[0], max_nontarget_lab_score_s[0]))
                    print("")
                    sys.stdout.flush()

                for batch_idx,(the_dist, the_score, the_adv_img, the_mask) in enumerate(zip(Loss_EN, OutputScore, adv_img, img_mask)):
                    if the_dist < current_step_best_dist[batch_idx] and compare(the_score, np.argmax(label_batch[batch_idx])):
                        current_step_best_dist[batch_idx] = the_dist
                        current_step_best_score[batch_idx] = np.argmax(the_score)
                    if the_dist < overall_best_dist[batch_idx] and compare(the_score, np.argmax(label_batch[batch_idx])):
                        overall_best_dist[batch_idx] = the_dist
                        overall_best_attack[batch_idx] = the_adv_img
                        overall_best_mask_vec[batch_idx] = img_mask

            # adjust the constant as needed
            for batch_idx in range(batch_size):
                if compare(current_step_best_score[batch_idx], np.argmax(label_batch[batch_idx])) and current_step_best_score[batch_idx] != -1:
                    # success, divide const by two
                    Const_UB[batch_idx] = min(Const_UB[batch_idx],CONST[batch_idx])
                    if Const_UB[batch_idx] < 1e9:
                        CONST[batch_idx] = (Const_LB[batch_idx] + Const_UB[batch_idx])/2
                else:
                    # failure, either multiply by 10 if no solution found yet
                    #          or do binary search with the known upper bound
                    Const_LB[batch_idx] = max(Const_LB[batch_idx],CONST[batch_idx])
                    if Const_UB[batch_idx] < 1e9:
                        CONST[batch_idx] = (Const_LB[batch_idx] + Const_UB[batch_idx])/2
                    else:
                        CONST[batch_idx] *= 10

        # return the best solution found
        overall_best_attack = overall_best_attack[0]
        overall_best_mask_vec = overall_best_mask_vec[0]
        # overall_best_mask_vec = overall_best_mask_vec.reshape(-1)
        return overall_best_attack.reshape((1,) + overall_best_attack.shape), overall_best_mask_vec

        """
        export PYTHONPATH=$PYTHONPATH:/u/pinyu/progressive_growing_of_gans
        python3 gen_example.py -s 6

        """
    
    def generate_PP(self, img_mask, orig_img, orig_class, model, mask_size, mask_mat):
        def model_prediction(model, inputs):
            prob = model.model.predict(inputs)
            predicted_class = np.argmax(prob)
            prob_str = np.array2string(prob).replace('\n','')
            return prob, predicted_class, prob_str        # ranking

        success = False
        print("Start ranking:")
        mask_vec = img_mask.reshape(-1)
        sort_idx = np.argsort(mask_vec)
        total_nonezero = len(np.argsort(mask_vec>0))
        working_mask = np.zeros((1,) + (mask_size, mask_size) + (1,))
        for i in range(1,total_nonezero):
            temp_index = sort_idx[-i]
            mask_position = np.argwhere(mask_mat[temp_index]==1)
            for index in mask_position:
                working_mask[(0,) + tuple(index) + (0,)] = 1
            adv_img = working_mask * orig_img
            img_prob, img_class, img_prob_str = model_prediction(model, adv_img)
            print("i:{}, index:{}, value:{}, class:{}".format(i, temp_index, mask_vec[temp_index], img_class))
            if img_class == orig_class:
                success = True
                break
                
        return adv_img, success



import numpy as np
import os


class CelebADataset():
    """
    Images are based on the CelebA Dataset [#1]_ [#2]_. Specifically, we use a GAN developed by Karras et. al [#3]_ in order to generate
    new images similar to CelebA. We use these generated images in order to also store the latent variables used to generate
    them, which are required for generating pertinent negatives in CEM-MAF [#4]_.

    References:
        .. [#1] `Liu, Luo, Wang, Tang. Large-scale CelebFaces Attributes (CelebA) Dataset. <http://mmlab.ie.cuhk.edu.hk/projects/CelebA.html>`_
        .. [#2] `Liu, Luo, Wang, Tang. Deep Learning Face Attributes in the Wild. ICCV. 2015. <https://arxiv.org/pdf/1411.7766.pdf>`_
        .. [#3] `Karras, Aila, Laine, Lehtinen Progressive Growing of GANs for Improved Quality, Stability, and Variation. ICLR. 2018.
           <https://github.com/tkarras/progressive_growing_of_gans>`_
        .. [#4] `Luss, Chen, Dhurandhar, Sattigeri, Shanmugam, Tu. Generating Contrastive Explanations with Monotonic Attribute Functions.
           2019. <https://arxiv.org/abs/1905.12698>`_
    """
    def __init__(self, dirpath=None):
        self._dirpath = dirpath
        if not self._dirpath:
            self._dirpath = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                '..', 'data','celeba_data')

    def get_img(self, img_id):
        return np.load(os.path.join(self._dirpath, "{}_img.npy".format(img_id)))

    def get_latent(self, img_id):
        return np.load(os.path.join(self._dirpath, "{}_latent.npy".format(img_id))).astype("float32")



import abc
import sys

# Ensure compatibility with Python 2/3 
if sys.version_info >= (3, 4):
    ABC = abc.ABC
else:
    ABC = abc.ABCMeta(str('ABC'), (), {})


class LocalBBExplainer(ABC):
        
    """ 
    LocalBBExplainer is the base class for local post-hoc black-box explainers (LBBE).
    Such explainers are model agnostic and generally require access to model's predict function alone.
    Examples include LIME[#1]_, SHAP[#2]_, etc..

    References:
        .. [#1] “Why Should I Trust You?” Explaining the Predictions of Any Classifier, ACM SIGKDD 2016.
        Marco Tulio Ribeiro, Sameer Singh, Carlos Guestrin. https://arxiv.org/abs/1602.04938.
        .. [#2] A Unified Approach to Interpreting Model Predictions, NIPS 2017.
        Lundberg, Scott M and Lee, Su-In. https://arxiv.org/abs/1705.07874

    """ 
    def __init__(self, *argv, **kwargs):
        
        """
        Initialize a LocalBBExplainer object. 
        ToDo: check common steps that need to be distilled here. 
        """
                
    @abc.abstractmethod 
    def set_params(self, *argv, **kwargs): 
        """
        Set parameters for the explainer. 
        """
        raise NotImplementedError 
        
    
    @abc.abstractmethod 
    def explain_instance(self, *argv, **kwargs):
        """
        Explain an input instance x.
        """
        raise NotImplementedError 



import abc
import sys

# Ensure compatibility with Python 2/3
if sys.version_info >= (3, 4):
    ABC = abc.ABC
else:
    ABC = abc.ABCMeta(str('ABC'), (), {})


class GlobalWBExplainer(ABC):
    """
    GlobalWBExplainer is the base class for Global white-box explainers (DIE).
    Given a source model, such explainers generally train a surrogate that is explainable.
    Examples include ProfWt[#1], etc.

    References:
        .. [#1] Improving Simple Models with Confidence Profiles (ProfWt),
        NIPS 2018, Amit Dhurandhar, Karthikeyan Shanmugam, Ronny Luss, Peder Olsen.
        https://arxiv.org/abs/1807.07506
    """

    def __init__(self, *argv, **kwargs):
        """
        Initialize a GlobalWBExplainer object.
        ToDo: check common steps that need to be distilled here.
        """

    @abc.abstractmethod
    def set_params(self, *argv, **kwargs):
        """
        Set parameters for the explainer.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def explain(self, *argv, **kwargs):
        """
        Explain model.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def fit(self, *argv, **kwargs):
        """
        Train a surrogate model.
        """
        raise NotImplementedError



from __future__ import print_function

from aix360.algorithms.lwbe import LocalWBExplainer 
from aix360.algorithms.contrastive.CEM_MAF_aen_PN import AEADEN as AEADEN_PN
from aix360.algorithms.contrastive.CEM_MAF_aen_PP import AEADEN as AEADEN_PP
from tensorflow.contrib.keras.api.keras.models import model_from_json

import os
import sys
import random
import time
import numpy as np
from skimage.segmentation import slic


class CEM_MAFImageExplainer(LocalWBExplainer):
    """CEM_MAFImageExplainer is a Contrastive Image explainer that leverages Monotonic
    Attribute Functions. The main idea here is to explain images using high level semantically meaningful attributes 
    that may either be directly available or learned through supervised or unsupervised methods. [#]_

    References:
        .. [#] `Ronny Luss, Pin-Yu Chen, Amit Dhurandhar, Prasanna Sattigeri,
           Karthikeyan Shanmugam, Chun-Chen Tu, "Generating Contrastive
           Explanations with Monotonic Attribute Functions," 2019.
           <https://arxiv.org/abs/1905.12698>`_
    """
    def __init__(self, model, attributes, aix360_path):
        """Initialize image explainer.

        Currently accepting model input which is an ImageClassifier.
        """
        super(CEM_MAFImageExplainer, self).__init__()
        self._wbmodel = model
        self._attributes = attributes
        self._aix360_path = aix360_path

    def set_params(self, *argv, **kwargs):
        """Set parameters for the explainer."""
        pass

    def explain_instance(self, sess, input_img, input_latent, arg_mode, arg_kappa, arg_binary_search_steps,
                        arg_max_iterations, arg_initial_const, arg_gamma, arg_beta, arg_attr_reg=1,
                        arg_attr_penalty_reg=1, arg_latent_square_loss_reg=1):
        """Explains an input instance input_image e.g. celebA is shape (1, 224, 224, 3)

        Hard coded batch_size=1, assuming we provide explanation for 1 input_image at a time. Returns
        either pertinent positive or pertinent depending on parameter.

        Args:
            sess (tensorflow.python.client.session.Session): Tensorflow session
            input_img (numpy.ndarray): image to be explained, of shape (1, size, size, channels)
            input_latent (numpy.ndarray): image to be explained, of shape (1, size, size, channels)
                in the latent space
            arg_mode (str): "PN" for pertinent negative or "PP" for pertinent positive
            arg_kappa (float): Confidence parameter that controls difference between prediction of
                PN (or PP) and original prediction
            arg_binary_search_steps (int): Controls number of random restarts to find best PN or PP
            arg_max_iterations (int): Max number iterations to run some version of gradient descent on
                PN or PP optimization problem from a single random initialization, i.e., total
                number of iterations wll be arg_binary_search_steps * arg_max_iterations
            arg_initial_const (int): Constant used for upper/lower bounds in binary search
            arg_gamma (float): Penalty parameter encouraging addition of attributes for PN or PP
            arg_beta (float): Penalty parameter encourages minimal addition of attributes to PN
                or sparsity of the mask that generates the PP
            arg_attr_reg (float): Penalty parameter on regularization of PN to be predicted different from
                original image
            arg_attr_penalty_reg (float): Penalty regularizing PN from being too different from original image
            arg_latent_square_loss_reg (float): Penalty regularizing PN from being too different from original
                image in the latent space

        Returns:
            tuple:
                * **adv_img** (`numpy.ndarray`) -- the pertinent positive or the pertinent negative image
                * **attr_mod** (`str`) -- only for PN; a string detailing which attributes were modified from the
                  original image
                * **INFO** (`str`) -- only for PN; a string of information about original vs PN class and
                  original vs PN prediction probability
        """

        random.seed(121)
        np.random.seed(1211)

        # %%change%%
        #(orig_prob, orig_class, orig_prob_str) = util.model_prediction(model, input_img)
        (orig_prob, orig_class, orig_prob_str) = self._wbmodel.predict_long(input_img)

        if arg_mode == 'PN':
            target_label = [np.eye(self._wbmodel._nb_classes)[orig_class]]


            attack_pn = AEADEN_PN(sess, self._wbmodel, attributes=self._attributes, aix360_path=self._aix360_path,
                            mode = arg_mode, batch_size=1, kappa=arg_kappa, init_learning_rate=1e-2,
                            binary_search_steps=arg_binary_search_steps, max_iterations=arg_max_iterations,
                            initial_const=arg_initial_const, gamma=arg_gamma, attr_reg=arg_attr_reg,
                            attr_penalty_reg=arg_attr_penalty_reg, latent_square_loss_reg=arg_latent_square_loss_reg)

            adv_img = attack_pn.attack(input_img, target_label, input_latent)
            adv_prob, adv_class, adv_prob_str = self._wbmodel.predict_long(adv_img)
            attr_mod = self.check_attributes_celebA(self._attributes, input_img, adv_img)
            
            INFO = "[INFO] Orig class:{}, Adv class:{}, Orig prob:{}, Adv prob:{}".format(orig_class, adv_class, orig_prob_str, adv_prob_str)

        else: # assume arg_mode is PP
            print("Creating a mask for pertinent positive")
            # create mask
            arg_seg_number = 200
            # Segment the original image using and create a mask for the segmentation
            #data = CELEBA_wrapper(os.path.join(img_path, "{}_img.npy".format(img_id)), orig_class, model)
            mask_label = slic(input_img, n_segments=arg_seg_number)[0]
            mask_num = len(np.unique(mask_label))
            mask_size = mask_label.shape[0]
            mask_mat = np.zeros((mask_num, mask_size, mask_size))
            for i in range(mask_num):
                temp_idx = np.argwhere(mask_label==i)
                for j in temp_idx:
                    mask_mat[(i,) + tuple(j)] = 1

            attack_pp = AEADEN_PP(sess, self._wbmodel, mask_mat=mask_mat, mode=arg_mode, batch_size=1, \
                                kappa=arg_kappa, init_learning_rate=1e-2, binary_search_steps=arg_binary_search_steps, \
                                max_iterations=arg_max_iterations, initial_const=arg_initial_const, beta=arg_beta, \
                                gamma=arg_gamma, attributes=self._attributes, aix360_path=self._aix360_path)


            target = np.zeros(self._wbmodel._nb_classes)
            target[orig_class]=1
            adv_img, img_mask = attack_pp.attack(input_img, [target])
            adv_prob, adv_class, adv_prob_str = self._wbmodel.predict_long(adv_img)

            print('Generating the pertinent positive')
            # Generate the PP
            success = False
            print("Start ranking:")
            mask_vec = img_mask.reshape(-1)
            sort_idx = np.argsort(mask_vec)
            total_nonezero = len(np.argsort(mask_vec>0))
            working_mask = np.zeros((1,) + (mask_size, mask_size) + (1,))
            for i in range(1,total_nonezero):
                temp_index = sort_idx[-i]
                mask_position = np.argwhere(mask_mat[temp_index]==1)
                for index in mask_position:
                    working_mask[(0,) + tuple(index) + (0,)] = 1
                adv_img = working_mask * input_img
                img_prob, img_class, img_prob_str = self._wbmodel.predict_long(adv_img)
                print("i:{}, index:{}, value:{}, class:{}".format(i, temp_index, mask_vec[temp_index], img_class))
                if img_class == orig_class:
                    success = True
                    break

            attr_mod = None
            INFO = None

        return(adv_img, attr_mod, INFO)

      
    def check_attributes_celebA(self, attributes, x, y): 
        """
        Load attribute classifiers and check which attributes in original image x
        are modified in adversarial image y

        Args:
            attributes (str list): list of attributes to load attribute classifiers for
            x (numpy.ndarray): original image
            y (numpy.ndarray): adversarial image

        Returns:
            str: string detailing which attributes were added to (or removed from)
            x resulting in y
        """

        orig_attr_score = np.zeros((len(attributes),1))
        adv_attr_score = np.zeros((len(attributes),1))
        for i in range(len(attributes)):
            attr = attributes[i]
            # load json and create model
            json_file_name = "../../aix360/models/CEM_MAF/simple_{}_model.json".format(attr)
            json_file = open(json_file_name, 'r')
            loaded_model_json = json_file.read()
            json_file.close()
            loaded_model = model_from_json(loaded_model_json)
            # load weights into new model
            weight_file_name = "../../aix360/models/CEM_MAF/simple_{}_weights.h5".format(attr)
            loaded_model.load_weights(weight_file_name)

            orig_attr_score[i] = loaded_model.predict(x)[0]
            adv_attr_score[i] = loaded_model.predict(y)[0]

        # pre-determined thresholds for changes in prediction values
        thresh_pos = np.zeros((len(attributes),1))
        thresh_pos[0] = .15
        thresh_pos[1] = .15
        thresh_pos[2] = .15
        thresh_pos[3] = .15
        thresh_pos[4] = .15
        thresh_pos[5] = .15
        thresh_pos[6] = .1
        thresh_pos[7] = .25
        thresh_pos[8] = .1
        thresh_pos[9] = .15
        thresh_pos[10] = .15
        thresh_pos[11] = .15

        thresh_neg = np.zeros((len(attributes),1))
        thresh_neg[0] = -.25
        thresh_neg[1] = -.25
        thresh_neg[2] = -.25
        thresh_neg[3] = -.25
        thresh_neg[4] = -.35
        thresh_neg[5] = -.25
        thresh_neg[6] = -.12
        thresh_neg[7] = -.25
        thresh_neg[8] = -.25
        thresh_neg[9] = -.25
        thresh_neg[10] = -.25
        thresh_neg[11] = -.25

        changes_abs = adv_attr_score - orig_attr_score
        changes = np.zeros((len(attributes),1))
        res = ""
        for i in range(len(attributes)):
            if changes_abs[i] >= thresh_pos[i]:
                changes[i] = 1
            elif changes_abs[i] <= thresh_neg[i]:
                changes[i] = -1
        added = np.where(changes == 1)[0]
        for j in range(len(added)):
            res += "Added " + attributes[added[j]] + ","
        removed = np.where(changes[i] == -1)[0]
        for j in range(len(removed)):
            res += "Removed " + attributes[removed[j]] + ","
        return res[:-1]



## Utils.py -- Some utility functions 
##
## Copyright (C) 2018, PaiShun Ting <paishun@umich.edu>
##                     Chun-Chen Tu <timtu@umich.edu>
##                     Pin-Yu Chen <Pin-Yu.Chen@ibm.com>
## Copyright (C) 2017, Huan Zhang <ecezhang@ucdavis.edu>.
## Copyright (C) 2016, Nicholas Carlini <nicholas@carlini.com>.
##
## This program is licenced under the BSD 2-Clause licence,
## contained in the LICENCE file in this directory.

import tensorflow as tf
import os
import numpy as np
import scipy.misc
from tensorflow.contrib.keras.api.keras.applications.resnet50 import ResNet50
from tensorflow.contrib.keras.api.keras.applications.vgg16 import VGG16
from tensorflow.contrib.keras.api.keras.layers import Input, Dense, Dropout, LeakyReLU, Activation
from tensorflow.contrib.keras.api.keras.models import Model, model_from_json, Sequential
from tensorflow.contrib.keras.api.keras.callbacks import ModelCheckpoint
from tensorflow.contrib.keras.api.keras import metrics
from tensorflow.contrib.keras.api.keras import regularizers
from tensorflow.contrib.keras.api.keras.optimizers import SGD

def load_AE(codec_prefix, print_summary=False):

    saveFilePrefix = "models/AE_codec/" + codec_prefix + "_"

    decoder_model_filename = saveFilePrefix + "decoder.json"
    decoder_weight_filename = saveFilePrefix + "decoder.h5"

    if not os.path.isfile(decoder_model_filename):
        raise Exception("The file for decoder model does not exist:{}".format(decoder_model_filename))
    json_file = open(decoder_model_filename, 'r')
    decoder = model_from_json(json_file.read(), custom_objects={"tf": tf})
    json_file.close()

    if not os.path.isfile(decoder_weight_filename):
        raise Exception("The file for decoder weights does not exist:{}".format(decoder_weight_filename))
    decoder.load_weights(decoder_weight_filename)

    if print_summary:
        print("Decoder summaries")
        decoder.summary()

    return decoder

class CELEBAModel:
    def __init__(self, nn_type="resnet50", restore = None, session=None, use_imagenet_pretrain=False, use_softmax=True):
        self.image_size = 224
        self.num_channels = 3
        self.num_labels = 8
    
        input_layer = Input(shape=(self.image_size, self.image_size, self.num_channels))
        weights = "imagenet" if use_imagenet_pretrain else None
        if nn_type == "resnet50":
            base_model = ResNet50(weights=weights, input_tensor=input_layer)
        elif nn_type == "vgg16":
            base_model = VGG16(weights=weights, input_tensor=input_layer)
            # base_model = VGG16(weights=None, input_tensor=input_layer)
        x = base_model.output
        x = LeakyReLU()(x)
        x = Dense(1024)(x)
        x = Dropout(0.2)(x)
        x = LeakyReLU()(x)
        x = Dropout(0.3)(x)
        x = Dense(8)(x)
        if use_softmax:
            x = Activation("softmax")(x)
        model = Model(inputs=base_model.input, outputs=x)
    
        # for layer in base_model.layers:
        # 	layer.trainable = False
    
    
        if restore:
            print("Load: {}".format(restore))
            model.load_weights(restore)
    
        self.model = model
    
    def predict(self, data):
        # this is used inside tf session, data should be a tensor
        return self.model(data)



import tensorflow as tf
import aix360.algorithms.profwt.resnet_keras_model
import numpy as np
from tensorflow.python.tools.inspect_checkpoint import print_tensors_in_checkpoint_file
import json



def print_layer_labels(checkpoint_path):
	"""
	Loads the tensorflow computation graph from the checkpoint
	Prints all operations names defined in the computation graph defined by the checkpoint.

	Parameters:
	checkpoint_path (string): Path to the checkpoint filename where the tensorflow checkpoint is stored.
	"""
	g= tf.train.import_meta_graph(checkpoint_path+'.meta')
	graph = tf.get_default_graph()
	print(graph.get_operations())
	tf.reset_default_graph()
	return

def attach_probe_eval(input_features_name,label_name,operation_name,x_train2,y_train2,checkpoint_path):
	"""
	Flattens a layer specificed by operation name and then evaluates the flattened layer on a dataset.
	
	Parameters:
	input_features_name (string): String that specifies the feature input placeholder tensor
	input_label_name (string): String that specifies the label placeholder tensor
	operation_name (string): String that specifies the operation at the end of which the layer outputs need to be probed. Has to be one of the values in the list
	                printed by the print_layer_labels function
	x_train2 (numpy array): training samples on which the the flattened layers needs to be evalauted on.
	y_train2 (numpy array): labels of training samples on which the flattened layer needs to be evaulated on. The function does not use this 
				information but is needed to supply values to the label placeholder in a keras/tensorflow model.

	Return:
	probe_tensor_val (numpy array)- Evaluated layer output matrix of dimensions (num of samples x flattened layer dimension).

	"""

	# Build the computation graph correpsonding to the trained model from the checkpoint
	g= tf.train.import_meta_graph(checkpoint_path+'.meta')
	graph = tf.get_default_graph()
	# Get the tensor corresponding to the operation
	a=graph.get_tensor_by_name(operation_name)
	#Get the placeholders for inputs and labels.
	input_x=graph.get_tensor_by_name(input_features_name)
	label_y=graph.get_tensor_by_name(label_name)
	print(a)
	# Stop any gradient flowing through it
	a_sg = tf.stop_gradient(a)
	#Flatten and obtain a probe output.
	shape = a_sg.get_shape().as_list()     # a list: [None, 9, 2]
	dim = np.prod(shape[1:])            # dim = prod(9,2) = 18
	probe = tf.reshape(a_sg, [-1, dim])
	print(probe)

	# Loads the weights in the computation graph and evaluate the probe tensor on all samples provided
	batch_size=500

	saver=tf.train.Saver()
	with tf.Session() as sess:      
		saver.restore(sess,checkpoint_path)
		num_examples=len(x_train2)
		i=0
		for offset in range(0, num_examples, batch_size):
			print (i)
			end = offset + batch_size
			x_t,y_t=x_train2[offset:end],y_train2[offset:end]
			if i==0:
				probe_list=sess.run(probe,feed_dict={input_x: x_t, label_y: y_t})
			else:
				l_dum=sess.run(probe,feed_dict={input_x: x_t, label_y: y_t})
				probe_list=np.append(np.array(probe_list),np.array(l_dum),axis=0)  
			i=i+1
	tf.reset_default_graph()
	#Save the array of flattend probe outputs
	probe_tensor_val=probe_list
	assert str(type(probe_tensor_val))=="<class 'numpy.ndarray'>"  
	assert probe_tensor_val.shape[0]==len(x_train2)
	print('probe shape',probe_tensor_val.shape)
	return probe_tensor_val






from .mnist_dataset import MNISTDataset
from .meps_dataset import MEPSDataset
from .heloc_dataset import HELOCDataset
from .celeba_dataset import CelebADataset
from .cdc_dataset import CDCDataset
from .cifar_dataset import CIFARDataset
from .ted_dataset import TEDDataset
from .fashion_mnist_dataset import FMnistDataset



from __future__ import print_function

import torch
from torch import __init__
from torch.optim import Adam

from .dipvae_utils import VAE, DIPVAE
from .dipvae_utils import plot_reconstructions, plot_latent_traversal
from aix360.algorithms.die import DIExplainer

import os
import sys
import random
import time
import numpy as np


class DIPVAEExplainer(DIExplainer):
    """DIPVAEExplainer can be used to visualize the changes in the latent space of Disentangled Inferred Prior-VAE
    or DIPVAE [#1]_. This model is a Variational Autoencoder [#2]_ variant that leads to a
    disentangled latent space. This is achieved by matching the covariance of the prior distributions with the
    inferred prior.

    References:
        .. [#1] `Variational Inference of Disentangled Latent Concepts from Unlabeled Observations (DIP-VAE), ICLR 2018.
         Kumar, Sattigeri, Balakrishnan. <https://arxiv.org/abs/1711.00848>`_
        .. [#2] `Diederik P Kingma and Max Welling. Auto-encoding variational Bayes. ICLR, 2014. <https://arxiv.org/pdf/1312.6114.pdf>`_

    """
    def __init__(self, model_args, dataset=None, net=None, cuda_available=None):
        """
        Initialize DIPVAEExplainer explainer.

        Args:
            model_args: This should contain all the parameter required for the generative model training and
                inference. This includes model type (vae, dipvae-i, dipvae-ii, user-defined). The user-defined model can be
                passed to the parameter net of the fit() function. Each of the model should have encoder and decode function
                defined. See the notebook example for other model specific parameters.
            dataset: The dataset object.
            net: If not None this is the user specified generative model.
            cuda_available: If True use GPU.
        """

        super(DIPVAEExplainer, self).__init__()
        self.model_args = model_args

        torch.manual_seed(self.model_args.seed)

        if net is None:
            if self.model_args.model == "vae":
                self.net = VAE(num_nodes=self.model_args.num_nodes, activation_type=self.model_args.activation_type,
                                 latent_dim=self.model_args.latent_dim,
                                 op_dim=np.prod(dataset.data_dims), args=self.model_args, cuda_available=cuda_available)
            elif self.model_args.model == "dipvae-i":
                self.net = DIPVAE(num_nodes=self.model_args.num_nodes, activation_type=self.model_args.activation_type,
                                    latent_dim=self.model_args.latent_dim,
                                    op_dim=np.prod(dataset.data_dims), args=self.model_args, cuda_available=cuda_available, mode='i',
                                    output_activation_type=dataset.output_activation_type, likelihood_type=dataset.likelihood_type)
            elif self.model_args.model == "dipvae-ii":
                self.net = DIPVAE(num_nodes=self.model_args.num_nodes, activation_type=self.model_args.activation_type,
                                    latent_dim=self.model_args.latent_dim,
                                    op_dim=np.prod(dataset.data_dims), args=self.model_args, cuda_available=cuda_available, mode='ii',
                                    output_activation_type=dataset.output_activation_type, likelihood_type=dataset.likelihood_type)
        else:
            self.net = net

        self.cuda_available = cuda_available
        self.dataset = dataset

        if self.cuda_available:
            self.net = self.net.cuda()

    def set_params(self, *argv, **kwargs):
        """
        Set parameters for the explainer.
        """
        print("TBD: Implement set params in DIPVAEExplainer")

    def explain(self,
                input_images,
                edit_dim_id,
                edit_dim_value,
                edit_z_sample=False):

        """
        Edits the images in the latent space and returns the generated images.

        Args:
            input_images: The input images.
            edit_dim_id: The latent dimension id that need to be edited.
            edit_dim_value: The value that is assigned to the latent dimension with id edit_dim_id.
            edit_z_sample: If True will use the sample from encoder instead of the mean.
        Returns:
            Edited images.
        """

        reference_z, reference_mu, reference_std = self.net.encode(input_images)
        if edit_z_sample:
            edited_z = reference_z
            edited_z[:,edit_dim_id] = edit_dim_value
        else:
            edited_z = reference_mu
            edited_z[:, edit_dim_id] = edit_dim_value

        edited_images = self.net.decode(edited_z)

        return edited_images

    def fit(self, visualize=False, save_dir="results"):
        """
        Train the underlying generative model.

        Args:
            visualize: Plot reconstructions during fit.
            save_dir: directory where plots and model will be saved.
        Returns:
            elbo
        """

        torch.manual_seed(self.model_args.seed)

        optimizer = Adam(self.net.parameters(), lr=self.model_args.step_size)

        loss_epoch_list = []

        for epoch in np.arange(self.model_args.num_epochs):

            loss_epoch = 0.
            batch_id = 0

            for x, y in self.dataset.next_batch():
                #x, y = torch.tensor(x), torch.tensor(y)
                if self.cuda_available:
                    x = x.cuda()
                    y = y.cuda()

                # forward
                if "mnist" in self.dataset.name:
                    loss = self.net.neg_elbo(x.squeeze().view(-1, np.prod(self.dataset.data_dims)))
                else:
                    loss = self.net.neg_elbo(x)
                # backward
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                loss_epoch += loss

                batch_id += 1

                if visualize and batch_id % 10 == 0:


                    if not os.path.isdir(save_dir):
                        os.makedirs(save_dir)

                    plot_reconstructions(self.dataset, self.net, x, image_id_to_plot=2, epoch=epoch,
                                         batch_id = batch_id, save_dir=save_dir)
                    if batch_id % 100 == 0:
                        plot_latent_traversal(self, x, self.model_args, self.dataset, image_id_to_plot=2, epoch=epoch,
                                              batch_id = batch_id, save_dir=save_dir)
                        torch.save(self.net, os.path.join(save_dir, 'net.p'))

            loss_epoch_list.append(-loss_epoch / self.dataset.num_training_instances)
            print("Epoch {0} | ELBO {1}".format(epoch, -loss_epoch / self.dataset.num_training_instances))



        return loss_epoch_list



import sys
import numpy as np
import tensorflow as tf
import keras.backend as K


class AEADEN:

    def __init__(self, model, shape, mode, AE, batch_size, kappa, init_learning_rate,
                 binary_search_steps, max_iterations, initial_const, beta, gamma):
        
        """
        Constructor method.
        
        Args:
            model: KerasClassifier classification model
            arg_mode(str): 'PP' or 'PN'
            AE: Auto-encoder model
            batch_size(int): Number of samples in a batch
            kappa(double): Confidence gap between desired class and other classes
            init_learning_rate(double): Initial learning rate
            binary_search_steps(int): Number of search steps
            max_iterations(int): For each weighting of loss function number of iterations to search
            initial_const(double): Initial weighting of loss function
            beta (double): Weighting of L1 loss
            gamma (double): Weighting of auto-encoder
        """

        num_classes = model._nb_classes
        tf_sum = list(range(1, len(shape)))

        self.sess = K.get_session()

        self.INIT_LEARNING_RATE = init_learning_rate
        self.MAX_ITERATIONS = max_iterations
        self.BINARY_SEARCH_STEPS = binary_search_steps
        self.kappa = kappa
        self.init_const = initial_const
        self.batch_size = batch_size
        self.AE = AE
        self.mode = mode
        self.beta = beta
        self.gamma = gamma

        # these are variables to be more efficient in sending data to tf
        self.orig_img = tf.Variable(np.zeros(shape), dtype=tf.float32)
        self.adv_img = tf.Variable(np.zeros(shape), dtype=tf.float32)
        self.adv_img_s = tf.Variable(np.zeros(shape), dtype=tf.float32)
        self.target_lab = tf.Variable(np.zeros((batch_size, num_classes)), dtype=tf.float32)
        self.const = tf.Variable(np.zeros(batch_size), dtype=tf.float32)
        self.global_step = tf.Variable(0.0, trainable=False)

        # and here's what we use to assign them
        self.assign_orig_img = tf.placeholder(tf.float32, shape)
        self.assign_adv_img = tf.placeholder(tf.float32, shape)
        self.assign_adv_img_s = tf.placeholder(tf.float32, shape)
        self.assign_target_lab = tf.placeholder(tf.float32, (batch_size, num_classes))
        self.assign_const = tf.placeholder(tf.float32, [batch_size])

        """Fast Iterative Soft Thresholding"""
        """--------------------------------"""
        self.zt = tf.divide(self.global_step, self.global_step + tf.cast(3, tf.float32))

        cond1 = tf.cast(tf.greater(tf.subtract(self.adv_img_s, self.orig_img), self.beta), tf.float32)
        cond2 = tf.cast(tf.less_equal(tf.abs(tf.subtract(self.adv_img_s, self.orig_img)), self.beta), tf.float32)
        cond3 = tf.cast(tf.less(tf.subtract(self.adv_img_s, self.orig_img), tf.negative(self.beta)), tf.float32)
        upper = tf.minimum(tf.subtract(self.adv_img_s, self.beta), tf.cast(0.5, tf.float32))
        lower = tf.maximum(tf.add(self.adv_img_s, self.beta), tf.cast(-0.5, tf.float32))
        self.assign_adv_img = tf.multiply(cond1, upper) + tf.multiply(cond2, self.orig_img) + tf.multiply(cond3, lower)

        cond4 = tf.cast(tf.greater(tf.subtract(self.assign_adv_img, self.orig_img), 0), tf.float32)
        cond5 = tf.cast(tf.less_equal(tf.subtract(self.assign_adv_img, self.orig_img), 0), tf.float32)
        if self.mode == "PP":
            self.assign_adv_img = tf.multiply(cond5, self.assign_adv_img) + tf.multiply(cond4, self.orig_img)
        elif self.mode == "PN":
            self.assign_adv_img = tf.multiply(cond4, self.assign_adv_img) + tf.multiply(cond5, self.orig_img)

        self.assign_adv_img_s = self.assign_adv_img + tf.multiply(self.zt, self.assign_adv_img - self.adv_img)
        cond6 = tf.cast(tf.greater(tf.subtract(self.assign_adv_img_s, self.orig_img), 0), tf.float32)
        cond7 = tf.cast(tf.less_equal(tf.subtract(self.assign_adv_img_s, self.orig_img), 0), tf.float32)
        if self.mode == "PP":
            self.assign_adv_img_s = tf.multiply(cond7, self.assign_adv_img_s) + tf.multiply(cond6, self.orig_img)
        elif self.mode == "PN":
            self.assign_adv_img_s = tf.multiply(cond6, self.assign_adv_img_s) + tf.multiply(cond7, self.orig_img)

        self.adv_updater = tf.assign(self.adv_img, self.assign_adv_img)
        self.adv_updater_s = tf.assign(self.adv_img_s, self.assign_adv_img_s)

        """--------------------------------"""
        # prediction BEFORE-SOFTMAX of the model
        self.delta_img = self.orig_img - self.adv_img
        self.delta_img_s = self.orig_img - self.adv_img_s
        # %%change%%
        if self.mode == "PP":
            # self.ImgToEnforceLabel_Score = model.predict(self.delta_img)
            # self.ImgToEnforceLabel_Score_s = model.predict(self.delta_img_s)
            self.ImgToEnforceLabel_Score = model.predictsym(self.delta_img)
            self.ImgToEnforceLabel_Score_s = model.predictsym(self.delta_img_s)
        elif self.mode == "PN":
            # self.ImgToEnforceLabel_Score = model.predict(self.adv_img)
            # self.ImgToEnforceLabel_Score_s = model.predict(self.adv_img_s)
            self.ImgToEnforceLabel_Score = model.predictsym(self.adv_img)
            self.ImgToEnforceLabel_Score_s = model.predictsym(self.adv_img_s)

        # distance to the input data

        self.L2_dist = tf.reduce_sum(tf.square(self.delta_img), axis=tf_sum)
        self.L2_dist_s = tf.reduce_sum(tf.square(self.delta_img_s), axis=tf_sum)
        self.L1_dist = tf.reduce_sum(tf.abs(self.delta_img), axis=tf_sum)
        self.L1_dist_s = tf.reduce_sum(tf.abs(self.delta_img_s), axis=tf_sum)

        self.EN_dist = self.L2_dist + tf.multiply(self.L1_dist, self.beta)
        self.EN_dist_s = self.L2_dist_s + tf.multiply(self.L1_dist_s, self.beta)

        # compute the probability of the label class versus the maximum other
        self.target_lab_score = tf.reduce_sum((self.target_lab) * self.ImgToEnforceLabel_Score, 1)
        target_lab_score_s = tf.reduce_sum((self.target_lab) * self.ImgToEnforceLabel_Score_s, 1)
        self.max_nontarget_lab_score = tf.reduce_max((1 - self.target_lab) * self.ImgToEnforceLabel_Score -
                                                     (self.target_lab * 10000), 1)
        max_nontarget_lab_score_s = tf.reduce_max((1 - self.target_lab) * self.ImgToEnforceLabel_Score_s -
                                                  (self.target_lab * 10000), 1)
        if self.mode == "PP":
            Loss_Attack = tf.maximum(0.0, self.max_nontarget_lab_score - self.target_lab_score + self.kappa)
            Loss_Attack_s = tf.maximum(0.0, max_nontarget_lab_score_s - target_lab_score_s + self.kappa)
        elif self.mode == "PN":
            Loss_Attack = tf.maximum(0.0, -self.max_nontarget_lab_score + self.target_lab_score + self.kappa)
            Loss_Attack_s = tf.maximum(0.0, -max_nontarget_lab_score_s + target_lab_score_s + self.kappa)
        # sum up the losses
        self.Loss_L1Dist = tf.reduce_sum(self.L1_dist)
        self.Loss_L1Dist_s = tf.reduce_sum(self.L1_dist_s)
        self.Loss_L2Dist = tf.reduce_sum(self.L2_dist)
        self.Loss_L2Dist_s = tf.reduce_sum(self.L2_dist_s)
        self.Loss_Attack = tf.reduce_sum(self.const * Loss_Attack)
        self.Loss_Attack_s = tf.reduce_sum(self.const * Loss_Attack_s)

        if self.mode == "PP" and callable(self.AE):
            self.Loss_AE_Dist = self.gamma * tf.square(tf.norm(self.AE(self.delta_img) - self.delta_img))
            self.Loss_AE_Dist_s = self.gamma * tf.square(tf.norm(self.AE(self.delta_img) - self.delta_img_s))
        elif self.mode == "PN" and callable(self.AE):
            self.Loss_AE_Dist = self.gamma * tf.square(tf.norm(self.AE(self.adv_img) - self.adv_img))
            self.Loss_AE_Dist_s = self.gamma * tf.square(tf.norm(self.AE(self.adv_img_s) - self.adv_img_s))
        else:
            self.Loss_AE_Dist = tf.constant(0.)
            self.Loss_AE_Dist_s = tf.constant(0.)

        self.Loss_ToOptimize = self.Loss_Attack_s + self.Loss_L2Dist_s + self.Loss_AE_Dist_s
        self.Loss_Overall = self.Loss_Attack + self.Loss_L2Dist + self.Loss_AE_Dist + tf.multiply(self.beta,
                                                                                                  self.Loss_L1Dist)

        self.learning_rate = tf.train.polynomial_decay(self.INIT_LEARNING_RATE, self.global_step, self.MAX_ITERATIONS,
                                                       0,
                                                       power=0.5)
        optimizer = tf.train.GradientDescentOptimizer(self.learning_rate)
        start_vars = set(x.name for x in tf.global_variables())
        self.train = optimizer.minimize(self.Loss_ToOptimize, var_list=[self.adv_img_s], global_step=self.global_step)
        end_vars = tf.global_variables()
        new_vars = [x for x in end_vars if x.name not in start_vars]

        # these are the variables to initialize when we run
        self.setup = []
        self.setup.append(self.orig_img.assign(self.assign_orig_img))
        self.setup.append(self.target_lab.assign(self.assign_target_lab))
        self.setup.append(self.const.assign(self.assign_const))
        self.setup.append(self.adv_img.assign(self.assign_adv_img))
        self.setup.append(self.adv_img_s.assign(self.assign_adv_img_s))

        self.init = tf.variables_initializer(var_list=[self.global_step] + [self.adv_img_s] + [self.adv_img] + new_vars)

    def attack(self, imgs, labs):
        """
        Perturbations to generate explanations
        """
        
        def compare(x,y):
            """
            compare x and y
            """
            if not isinstance(x, (float, int, np.int64)):
                x = np.copy(x)
                # x[y] -= self.kappa if self.PP else -self.kappa
                if self.mode == "PP":
                    x[y] -= self.kappa
                elif self.mode == "PN":
                    x[y] += self.kappa
                x = np.argmax(x)
            if self.mode == "PP":
                return x==y
            else: 
                return x!=y
        
        batch_size = self.batch_size
                
        # set the lower and upper bounds accordingly
        Const_LB = np.zeros(batch_size)
        CONST = np.ones(batch_size)*self.init_const
        Const_UB = np.ones(batch_size)*1e10
        # the best l2, score, and image attack
        overall_best_dist = [1e10]*batch_size
        overall_best_attack = [np.zeros(imgs[0].shape)]*batch_size
        
        for binary_search_steps_idx in range(self.BINARY_SEARCH_STEPS):
            # completely reset adam's internal state.
            self.sess.run(self.init)
            img_batch = imgs[:batch_size]
            label_batch = labs[:batch_size]
            
            current_step_best_dist = [1e10]*batch_size
            current_step_best_score = [-1]*batch_size
            
            # set the variables so that we don't have to send them over again
            self.sess.run(self.setup, {self.assign_orig_img: img_batch,
                                       self.assign_target_lab: label_batch,
                                       self.assign_const: CONST,
                                       self.assign_adv_img: img_batch,
                                       self.assign_adv_img_s: img_batch})
            
            for iteration in range(self.MAX_ITERATIONS):
                # perform the attack
                self.sess.run([self.train])
                self.sess.run([self.adv_updater, self.adv_updater_s])
                
                Loss_Overall, Loss_EN, OutputScore, adv_img = self.sess.run([self.Loss_Overall, 
                                                                             self.EN_dist, self.ImgToEnforceLabel_Score,
                                                                             self.adv_img])
                
                Loss_Attack, Loss_L2Dist, Loss_L1Dist, Loss_AE_Dist = self.sess.run([self.Loss_Attack, self.Loss_L2Dist, 
                                                                                     self.Loss_L1Dist, self.Loss_AE_Dist])
                target_lab_score, max_nontarget_lab_score_s = self.sess.run([self.target_lab_score, 
                                                                             self.max_nontarget_lab_score])
                # %%change%% 
                if iteration%(self.MAX_ITERATIONS//2) == 0:
                    print("iter:{} const:{}". format(iteration, CONST))
                    print("Loss_Overall:{:.4f}, Loss_Attack:{:.4f}". format(Loss_Overall, Loss_Attack))
                    print("Loss_L2Dist:{:.4f}, Loss_L1Dist:{:.4f}, AE_loss:{}". format(Loss_L2Dist, Loss_L1Dist, Loss_AE_Dist))
                    print("target_lab_score:{:.4f}, max_nontarget_lab_score:{:.4f}". format(target_lab_score[0], 
                                                                                            max_nontarget_lab_score_s[0]))
                    print("")
                    sys.stdout.flush()
                    
                for batch_idx,(the_dist, the_score, the_adv_img) in enumerate(zip(Loss_EN, OutputScore, adv_img)):
                    if the_dist < current_step_best_dist[batch_idx] and compare(the_score, np.argmax(label_batch[batch_idx])):
                        current_step_best_dist[batch_idx] = the_dist
                        current_step_best_score[batch_idx] = np.argmax(the_score)
                    if the_dist < overall_best_dist[batch_idx] and compare(the_score, np.argmax(label_batch[batch_idx])):
                        overall_best_dist[batch_idx] = the_dist
                        overall_best_attack[batch_idx] = the_adv_img
                        
            # adjust the constant as needed
            for batch_idx in range(batch_size):
                if compare(current_step_best_score[batch_idx], np.argmax(label_batch[batch_idx])) and current_step_best_score[batch_idx] != -1:
                    # success, divide const by two
                    Const_UB[batch_idx] = min(Const_UB[batch_idx],CONST[batch_idx])
                    if Const_UB[batch_idx] < 1e9:
                        CONST[batch_idx] = (Const_LB[batch_idx] + Const_UB[batch_idx])/2
                else:
                    # failure, either multiply by 10 if no solution found yet
                    #          or do binary search with the known upper bound
                    Const_LB[batch_idx] = max(Const_LB[batch_idx],CONST[batch_idx])
                    if Const_UB[batch_idx] < 1e9:
                        CONST[batch_idx] = (Const_LB[batch_idx] + Const_UB[batch_idx])/2
                    else:
                        CONST[batch_idx] *= 10

        # return the best solution found
        overall_best_attack = overall_best_attack[0]
        return overall_best_attack.reshape((1,) + overall_best_attack.shape)

